//////////////////////////////////////////////////////////////////////////
// JJ Foote w/ parts by Michael Thornbury                               //
// PlaySPC_GTK                                                          //
// Version 0.2.7 - February 21, 2003                                    //
//                                                                      //
// This program is free software; you can redistribute it and/or modify //
// it under the terms of the GNU General Public License as published by //
// the Free Software Foundation; either version 2 of the License, or    //
// (at your option) any later version.                                  //
//                                                                      //
// This program is distributed in the hope that it will be useful,      //
// but WITHOUT ANY WARRANTY; without even the implied warranty of       //
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        //
// GNU General Public License for more details.                         //
//////////////////////////////////////////////////////////////////////////

#define CURRENT_VERSION "0.2.7"
#define MAX_BUFS 40
#define SHOUTCAST_CONNECTIONS_MAX 30
#define DEFAULT_BITRATE 64

#include "playspc_gtk.h"
#include <math.h>
#include "timer_macros.h"
#include "setting_macros.h"

#define USING_NONE 0
#define USING_RAR 1
#define USING_PLAYLIST 2
#define USING_CUSTOM 3

// Quick macro to handle writing to the input plugin.
#define write_cmd_handler(b) write_handler(b,COMMAND_SIZE)

// Quick macro to handle writing to the input plugin.
#define write_handler(b,c) \
while (write(input_pipe[STDOUT_FILENO],b,c) < 0) { \
if (errno != EAGAIN) { \
perror("Write error:"); \
clean_up_and_quit(1); \
} \
}

#define gtk_label_and_textfield(parent,textfield,title,callback,var) \
textfield = gtk_label_and_text_box(parent, title, var.integer); \
gtk_entry_set_text(GTK_ENTRY(textfield), var.string); \
gtk_signal_connect(GTK_OBJECT(textfield), "changed", GTK_SIGNAL_FUNC(callback), (gpointer) &var)


// A structure for our playlist entries. With this, it is now possible
// to discover the type of playlist used. So, if we end up having a 
// mixture of rar and custom or playlist, we can force the user to
// save as one or another type. Otherwise, we can default save much
// better.
typedef struct {
char full_filename[MAX_STRING_SIZE];
char source_filename[MAX_STRING_SIZE];
int fromrar;
} playlist_entry;

// This section is a rip-off of what icewm does.
#define ACOUNT(x) (sizeof(x)/sizeof(x[0]))

int OpenLinuxOut();
void Shutdown();

// Global variables
static gboolean true_option_id666, true_option_obey_length;
static gboolean now_playing, now_paused;
int using = USING_NONE, using_old;
//int alarm_pass;
//int alarm_count;
static playlist_entry playlistentry[MAX_PLAYLIST_ENTRIES];
static char filename[MAX_STRING_SIZE], old_filename[MAX_STRING_SIZE];
static char current_directory[MAX_STRING_SIZE];
static char temp_dir[] = TEMP_DIR "playspc-XXXXXX";
static int playlist_length = 0;
static int freq = FREQ;
static int buffer_size = FREQ*CHANNELS/URATE*sizeof(short);
static int buffer_rate = FREQ*CHANNELS/URATE;
static int current_row;
static int current_song_length;
static GtkWidget *playlist, *filename_textbox;
static TIMER_PAIR mytimer;
//static struct id666 new_id666_tag;
static struct id666 tag;
static int input_pipe[2];
static int encoder_pipe[2];
static int current_time;
static int total_time;
static int linuxout = 0;
static int shoutcast = 0;
static int shoutcast_connections = 0;
int map[SHOUTCAST_CONNECTIONS_MAX];
int connections;
char connections_name[40];
int queue_pos[SHOUTCAST_CONNECTIONS_MAX];
int queue_last_pos = 0;
int thread_done[SHOUTCAST_CONNECTIONS_MAX];
char* encode_buffers[MAX_BUFS];
int buf_size = DEFAULT_BITRATE*1024/8;
static int port = 8000;
static int encoder_pid = 0;
static int plugin_pid = 0;
static int bitrate = DEFAULT_BITRATE;
static int listener = 0;
static int stereo = 0;
static int echo = 0;
int limit = 0;
int lowerbound = 0, upperbound = 0, empty_space = 0;
int boost = 0;
int new_id;
pthread_mutex_t shoutcast_mutex; // Used to handling of shoutcast data.
pthread_cond_t shoutcast_cond; // Used to handling of shoutcast data.

//pthread_mutex_t alarm_mutex;

pthread_mutex_t clean_up_mutex;
// Used to guarantee these are only done once.
static int save_rc_done = 0, remove_temps_done = 0, gtk_done = 0;

struct option_gboolean_type {
const char* short_name;
const char* long_name;
const char* description;
gboolean* variable;
} option_gboolean[] = {
{ "use_id666_tags", "Use ID666 Tags", "Many SPC files contain an ID666 tag that describes the song's name, length, author, etc.  "
"This option enables parsing those tags.", &true_option_id666 },
{ "obey_length", "Obey Length in ID666 Tags (if present)", "This has no effect unless 'Use ID666 Tags' is enabled."
"If selected, an SPC will only play for as long as its ID666 header specifies (if it has one).", &true_option_obey_length }
};

struct option_int_type {
const char* short_name;
const char* long_name;
const char* description;
const int min;
const int max;
int* variable;
} option_int[] = {
{ "lowerbound", "Lowerbound", "Specifies the lower bound in seconds for a song to play if it does not have its own defined. (0 == UNLIMITED)", 0, 9999, &lowerbound },
{ "upperbound", "Upperbound", "Specifies the upper bound in seconds for a song to play for those songs which have a defined time. (0 == UNLIMITED)", 0, 9999, &upperbound },
{ "empty_space", "Empty time", "Specifies the number of seconds of no sound to wait for to switch to the next song. (0 == NEVER)", 0, 99, &empty_space }
};

struct option_string_type {
const char* short_name;
const char* long_name;
const char* description;
const int max_length;
char* variable;
} option_string[] = {
{ "home_dir", NULL, NULL, MAX_STRING_SIZE, current_directory }
};

struct char_and_int {
char* string;
int integer;
};

struct char_and_int filename_container = { filename, MAX_STRING_SIZE };
struct char_and_int tag_gametitle = { tag.gametitle, 33 };
struct char_and_int tag_songtitle = { tag.songtitle, 33};
struct char_and_int tag_songauthor = { tag.songauthor, 33};
struct char_and_int tag_songlength = { NULL, 5};
struct char_and_int tag_dumpername = { tag.dumpername, 17};
struct char_and_int tag_dumpdate = { tag.dumpdate, 11};
struct char_and_int tag_dumpprog = { tag.dumpprog, 7};
struct char_and_int tag_comments = { tag.comments, 33 };

static gboolean temp_gboolean[ACOUNT(option_gboolean)];
static char* temp_string[ACOUNT(option_string)];
static int temp_int[ACOUNT(option_int)];

// Load preferences from .rc file
gboolean load_rc()
{
FILE* RCFILE;
char tempstr[MAX_STRING_SIZE];
char temp[MAX_STRING_SIZE], temp2[MAX_STRING_SIZE];
int i;

snprintf(tempstr, MAX_STRING_SIZE, "%s/.playspc_gtkrc", getenv("HOME"));
RCFILE = fopen(tempstr, "r");
if(RCFILE == NULL) {
g_print("Error, cannot open %s for reading.\n", tempstr);
return FALSE;
} else {
while(fgets(tempstr, MAX_STRING_SIZE, RCFILE) != NULL) {
if (sscanf(tempstr, " %s = %s\n", temp, temp2) == 2) {
for (i = 0; i < ACOUNT(option_gboolean); i++)
{
   if (strncmp(temp, option_gboolean[i].short_name, MAX_STRING_SIZE) == 0) {
      if (strncmp("false", temp2, 5) == 0)
	 *(option_gboolean[i].variable) = FALSE;
      else
	 *(option_gboolean[i].variable) = TRUE;
      break;
   }
}

for (i = 0; i < ACOUNT(option_string); i++)
{
   if (strncmp(temp, option_string[i].short_name, MAX_STRING_SIZE) == 0) {
      strncpy(option_string[i].variable, temp2, MAX_STRING_SIZE);
      break;
   }
}

for (i = 0; i < ACOUNT(option_int); i++)
{
   if (strncmp(temp, option_int[i].short_name, MAX_STRING_SIZE) == 0) {
      *(option_int[i].variable) = atoi(temp2);
      break;
   }
}
}
}
fclose(RCFILE);
}
chdir(current_directory);
return TRUE;
}

// Save preferences to .rc file
gboolean save_rc()
{
FILE* RCFILE;
char tempstr[MAX_STRING_SIZE];
int i;

snprintf(tempstr, MAX_STRING_SIZE, "%s/.playspc_gtkrc", getenv("HOME"));
RCFILE = fopen(tempstr, "w");
if(RCFILE == NULL) {
g_print("Error, cannot open %s for writing.\n", tempstr);
return FALSE;
} else {
for (i = 0; i < ACOUNT(option_gboolean); i++)
{
write_boolean(tempstr,option_gboolean[i].short_name,*(option_gboolean[i].variable));
fputs(tempstr, RCFILE);
}

for (i = 0; i < ACOUNT(option_string); i++)
{
write_string(tempstr,option_string[i].short_name,option_string[i].variable);
fputs(tempstr, RCFILE);
}

for (i = 0; i < ACOUNT(option_int); i++)
{
write_int(tempstr,option_int[i].short_name,*(option_int[i].variable));
fputs(tempstr, RCFILE);
}

fclose(RCFILE);
}
return TRUE;
}

gboolean is_spc_file(char* spcfile)
{
FILE* INFILE;
char tempstr[28];

INFILE = fopen(spcfile, "r");
if(INFILE != NULL) {
fgets(tempstr, 28, INFILE);
// Checks for the spc header.
if(strncmp(tempstr, "SNES-SPC700 Sound File Data", 27) == 0) {
fclose(INFILE);
return TRUE;
}
fclose(INFILE);
}
return FALSE;
}

// Function to parse ID666 tags in SPC files.  Based on v0.30 of the SPC file format.
gboolean read_id666_tag(char* spcfile, struct id666* tag)
{
FILE* INFILE;
char tempstr[MAX_STRING_SIZE];
int	dumpprogcode;

INFILE = fopen(spcfile, "r");
if(INFILE == NULL) {
g_print("2> Cannot open %s for reading.  Giving up.\n", spcfile);
} else {
fgets(tempstr, 28, INFILE);

// Anal retentive check, though is_spc_file should have cleared out this condition.
if(strncmp(tempstr, "SNES-SPC700 Sound File Data", 27) != 0) {
g_print("Error, %s does not appear to be a valid SPC file!\n", spcfile);
} else {
fseek(INFILE, 0x23, SEEK_SET);
fgets(tempstr, 1, INFILE);
/* Supposedly 0x23 == 27 means it doesn't contain id666 data, but from what I can tell,
that standard is not followed at all. It should be okay to leave it in here for those
few files which really are compliant.*/
if (tempstr[0] == 27)
return FALSE;

fseek(INFILE, 46L, SEEK_SET);
fgets(tag->songtitle, 78 - 46 + 1, INFILE);

fseek(INFILE, 78L, SEEK_SET);
fgets(tag->gametitle, 110 - 78 + 1, INFILE);

fseek(INFILE, 110L, SEEK_SET);
fgets(tag->dumpername, 126 - 110 + 1, INFILE);

fseek(INFILE, 126L, SEEK_SET);
fgets(tag->comments, 158 - 126 + 1, INFILE);

fseek(INFILE, 158L, SEEK_SET);
fgets(tag->dumpdate, 168 - 158 + 1, INFILE);

fseek(INFILE, 169L, SEEK_SET);
fgets(tempstr, 172 - 169 + 1, INFILE);
/* This is what Merdian Advanced does, so I guess it's okay to follow
   that example. */
tag->songlength_raw = atoi(tempstr) + 3;

// Currently we ignore the fadeout, but read it in anyway.
fseek(INFILE, 172L, SEEK_SET);
fgets(tempstr, 177 - 172 + 1, INFILE);
tag->fadeout = atoi(tempstr);

fseek(INFILE, 177L, SEEK_SET);
fgets(tag->songauthor, 209 - 177 + 1, INFILE);

fseek(INFILE, 210L, SEEK_SET);
dumpprogcode = fgetc(INFILE);

snprintf(tag->songlength, 7, "%d:%02d", tag->songlength_raw / 60
				  , tag->songlength_raw % 60);

fclose(INFILE);

if(dumpprogcode == 0)
strcpy(tag->dumpprog, "Unknown");
else if(dumpprogcode == 1)
strcpy(tag->dumpprog, "ZSNES");
else if(dumpprogcode == 2)
strcpy(tag->dumpprog, "SNES9X");
else
strcpy(tag->dumpprog, "INVALID");

/* The most important part of an id666 IMO, so the thing we should really check for. */
if (tag->songlength_raw != 3)
return TRUE;
else
return FALSE;
}
}

return FALSE;
}

/* Function to write back an altered ID666 tag to an SPC file.  This is problematic for files in a RAR archive.
WARNING: this function has not been heavily tested, and could possibly destroy your SPC files.  Backup first! */
gboolean write_id666_tag(char* spcfile, struct id666* tag)
{
FILE* OUTFILE;
char tempstr[MAX_STRING_SIZE];
int dumpprogcode;

OUTFILE = fopen(spcfile, "r+");
if(OUTFILE == NULL) {
g_print("Cannot open %s for writing.  Giving up.\n", spcfile);
} else {
fputs("SNES-SPC700 Sound File Data v0.30", OUTFILE);

fseek(OUTFILE, 46L, SEEK_SET);
fputs(tag->songtitle, OUTFILE);
fputc('\0', OUTFILE);

fseek(OUTFILE, 78L, SEEK_SET);
fputs(tag->gametitle, OUTFILE);
fputc('\0', OUTFILE);

fseek(OUTFILE, 110L, SEEK_SET);
fputs(tag->dumpername, OUTFILE);
fputc('\0', OUTFILE);

fseek(OUTFILE, 126L, SEEK_SET);
fputs(tag->comments, OUTFILE);
fputc('\0', OUTFILE);

fseek(OUTFILE, 158L, SEEK_SET);
fputs(tag->dumpdate, OUTFILE);
fputc('\0', OUTFILE);

fseek(OUTFILE, 169L, SEEK_SET);
// Inverts the add done when reading in.
snprintf(tempstr, 4, "%d", tag->songlength_raw - 3); // itoa()
// Writes out exactly three characters, which means:
// #\0? for single digit values, where ? is garbage
// ##\0 for double digit values.
// ### for triple digit values (the limit).
fputc(tempstr[0], OUTFILE);
fputc(tempstr[1], OUTFILE);
fputc(tempstr[2], OUTFILE);

fseek(OUTFILE, 173L, SEEK_SET);
sprintf(tempstr, "%04d", tag->fadeout); // itoa()
fputs(tempstr, OUTFILE);
fputc('\0', OUTFILE);

fseek(OUTFILE, 177L, SEEK_SET);
fputs(tag->songauthor, OUTFILE);
fputc('\0', OUTFILE);

fseek(OUTFILE, 210L, SEEK_SET);
if(strcasecmp(tag->dumpprog, "ZSNES") == 0)
dumpprogcode = 1;
else if(strcasecmp(tag->dumpprog, "SNES9X") == 0)
dumpprogcode = 2;
else
dumpprogcode = 0;
fputc(dumpprogcode, OUTFILE);

fclose(OUTFILE);
return TRUE;
}
return FALSE;
}

/* RAR Archive handler */
gboolean rar_extract(char* rarfile)
{
//XXX: If the filename of an SPC file inside a RAR archive contains a " (double-quote) character,
//     this code will probably blow up on it.
char cmdstr[MAX_STRING_SIZE];
struct stat playlist_status;
FILE* INFILE;
char tempstr[MAX_STRING_SIZE];
int i;
gchar* trash[1];
int playlist_length_start = playlist_length;

trash[0] = (gchar*) malloc(sizeof(gchar) * MAX_STRING_SIZE);

snprintf(tempstr, MAX_STRING_SIZE, "%s/playlist", temp_dir);
unlink(tempstr);

// List the contents of the RAR file, and grep out " Word...". Checking what is
// or isn't an spc is done later.
snprintf(cmdstr, MAX_STRING_SIZE, RAR RAR_LIST_FLAGS "\"%s\" | "
		      "%s%s ^\\ \\\\\\< | "
		      "%s%s\"s/^\\ *//\" "
		      " > %s/rarlist.txt",
		      rarfile,
		      GREP, GREP_FLAGS,
		      SED, SED_FLAGS,
		      temp_dir);
system(cmdstr);

// Extract all files with periods in them to our temporary directory.
snprintf(cmdstr, MAX_STRING_SIZE, "%s%s\"%s\"  \\*.* %s/",
		      RAR, RAR_EXTRACT_FLAGS, rarfile,
		      temp_dir);
system(cmdstr);

snprintf(tempstr, MAX_STRING_SIZE, "%s/playlist", temp_dir);

stat(tempstr, &playlist_status);

if (!S_ISREG(playlist_status.st_mode))
snprintf(tempstr, MAX_STRING_SIZE, "%s/rarlist.txt", temp_dir);

INFILE = fopen(tempstr, "r");
if(INFILE == NULL) {
g_print("3> Cannot open %s for reading.  Giving up.\n", tempstr);
strncpy(filename, old_filename, MAX_STRING_SIZE);
gtk_entry_set_text(GTK_ENTRY(filename_textbox), filename);
using = using_old;
//        free(trash[0]);
return FALSE;
} else {
while(fgets(tempstr, MAX_STRING_SIZE, INFILE) != NULL) {
// Strip off the newline that fgets might slap on the end of our string
for(i = 0; i < MAX_STRING_SIZE; i++) {
if(tempstr[i] == '\n') {
    tempstr[i] = '\0';
    break;
}
}

// Prepend the path to each SPC filename, using trash[0] as a temporary because we don't need it yet
snprintf(trash[0], MAX_STRING_SIZE, "%s/%s",
				temp_dir, tempstr);
snprintf(tempstr, MAX_STRING_SIZE, "%s", trash[0]);

// Only add real spc files to our list.
if (is_spc_file(tempstr)) {
strncpy(playlistentry[playlist_length].full_filename, tempstr, MAX_STRING_SIZE);
strncpy(playlistentry[playlist_length].source_filename, rarfile, MAX_STRING_SIZE);
playlistentry[playlist_length].fromrar = 1;
fill_name(trash[0],playlist_length);
gtk_clist_append(GTK_CLIST(playlist), trash);
playlist_length++;
}
}
if (playlist_length == playlist_length_start)
g_print("Error, %s does not appear to contain any SPC files!\n", rarfile);
fclose(INFILE);
}

using = USING_RAR;

//    free(trash[0]);
return TRUE;
}

// Fill in a name of either:
// for id666: gametitle - songtitle [(time_limit)]
// or else:   source_filename - basename(full_filename) [(time_limit)]

void fill_name(char* filled_name, int row)
{
struct id666 tag;
char* total;
char* song;
char scratch1[MAX_STRING_SIZE];
char scratch2[MAX_STRING_SIZE];
int time = 0;
int current_song_length;

strncpy(scratch1, playlistentry[row].source_filename, MAX_STRING_SIZE);
strncpy(scratch2, playlistentry[row].full_filename, MAX_STRING_SIZE);
strncpy(scratch1, basename(scratch1), MAX_STRING_SIZE);
strncpy(scratch2, basename(scratch2), MAX_STRING_SIZE);
total = scratch1;
song = scratch2;


if (true_option_id666 && true_option_obey_length && read_id666_tag(playlistentry[row].full_filename, &tag)) {
   current_song_length = tag.songlength_raw;
   if (current_song_length < 0)
      current_song_length = lowerbound;
   if (upperbound) {
      if (current_song_length > upperbound)
         time = upperbound;
      else
         time = current_song_length;
   } else {
     time = current_song_length;
   }
   if (tag.gametitle[0] != '\0')
      total = tag.gametitle;
   if (tag.songtitle[0] != '\0')
      song = tag.songtitle;
} else
   time = lowerbound;

if (time)
snprintf(filled_name, MAX_STRING_SIZE, "%i) %s - %s (%d:%02d)", row + 1, total, song, time/60, time%60);
else
snprintf(filled_name, MAX_STRING_SIZE, "%i) %s - %s", row + 1, total, song);
}

int fork_input_plugin(char* name)
{
// A child thread (not too important to store).
pthread_t child;
pthread_attr_t attr;

int tmp_pipe1[2];
int tmp_pipe2[2];

if (pipe(tmp_pipe1) || pipe(tmp_pipe2)) {
g_print("Error opening pipe to communicate with input player.\n");
return 1;
}

plugin_pid = fork();
if(plugin_pid == -1)
return 1;

if(plugin_pid == 0) {
dup2(tmp_pipe1[STDIN_FILENO], STDIN_FILENO);
dup2(tmp_pipe2[STDOUT_FILENO], STDOUT_FILENO);
if (execlp(name, name, NULL) == -1)
perror("Error starting plug-in:");
return 1;
}

close(tmp_pipe1[STDIN_FILENO]);
close(tmp_pipe2[STDOUT_FILENO]);
input_pipe[STDIN_FILENO] = tmp_pipe2[STDIN_FILENO];
input_pipe[STDOUT_FILENO] = tmp_pipe1[STDOUT_FILENO];

//    signal(SIGALRM, alarm_handler);

pthread_attr_init(&attr);
// Make it detached.
pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
// And one-to-one with kernel threads.
pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);

// Creates a seperate thread to handle input from the spc player.
pthread_create(&child, &attr, input_plugin_handler, NULL);

return 0;
}

// Some basic initialization
int init()
{
int i;

signal(SIGCHLD, SIG_IGN); // Ignore child processes, to avoid zombies

// Initialize mutexes.
pthread_mutex_init(&shoutcast_mutex, NULL);

pthread_mutex_init(&clean_up_mutex, NULL);

//    pthread_mutex_init(&alarm_mutex, NULL);

pthread_cond_init(&shoutcast_cond, NULL);

// Let each thread handle a signal pipe.
signal(SIGPIPE, SIG_IGN);

// Also handle termination of process.
signal(SIGTERM, clean_up_and_quit);
signal(SIGINT, clean_up_and_quit);

filename[0] = '\0';
now_playing = FALSE;

if(load_rc() == FALSE) {
true_option_id666 = TRUE;
true_option_obey_length = FALSE;
}

for (i = 0; i < ACOUNT(option_gboolean); i++)
temp_gboolean[i] = *(option_gboolean[i].variable);

for (i = 0; i < ACOUNT(option_int); i++)
temp_int[i] = *(option_int[i].variable);

for (i = 0; i < ACOUNT(option_string); i++) {
temp_string[i] = malloc(MAX_STRING_SIZE * sizeof(char));
strncpy(temp_string[i], option_string[i].variable, MAX_STRING_SIZE);
}

fork_input_plugin(PLUG_IN);

/* Get frequency. */
if (linuxout == 0) {
if (OpenLinuxOut()) {
g_print("Error: Unable to open sound device.\n");
return 1;
}
Shutdown();
}

return 0;
}

int init2()
{
int send_pipe[2];
int recv_pipe[2];
int i;
char freq_str[10], bitrate_str[10];

// A child thread (not too important to store).
pthread_t child, child2, child3;
pthread_attr_t attr;

if (shoutcast) {
for (i = 0; i < MAX_BUFS; i++)
encode_buffers[i] = malloc(buf_size);
for (i = 0; i < SHOUTCAST_CONNECTIONS_MAX; i++)
queue_pos[i] = -1;
pthread_attr_init(&attr);
// Make it detached.
pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
// And one-to-one with kernel threads.
pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);

for (i = 0; i < SHOUTCAST_CONNECTIONS_MAX; i++)
map[i] = -1;

// Creates a seperate thread for a shoutcast server.
pthread_create(&child, &attr, shoutcast_server, NULL);

if (pipe(send_pipe) || pipe(recv_pipe)) {
g_print("Error opening pipe to communicate with encoder.\n");
return 1;
}

encoder_pid = fork();
if(encoder_pid == -1)
return 1;

if (encoder_pid == 0) {
dup2(send_pipe[STDIN_FILENO], STDIN_FILENO);
dup2(recv_pipe[STDOUT_FILENO], STDOUT_FILENO);
snprintf(freq_str, 9, "%f", ((float) freq / (float) 1000));
snprintf(bitrate_str, 5, "%d", bitrate);
if (stereo) {
if (execlp(ENCODER, ENCODER, "-r", "--bitwidth", "16", "-x", "-m", "s",
	       "-b", bitrate_str, "-s",  freq_str, "-h",
	       "-", "-", NULL) == -1)
perror("Error starting encoder:");
} else {
if (execlp(ENCODER, ENCODER, "-r", "--bitwidth", "16", "-x", "-m", "m",
	      "-b", bitrate_str, "-s",  freq_str, "-h",
	      "-", "-", NULL) == -1)
perror("Error starting encoder:");
}
return 1;
}

close(send_pipe[STDIN_FILENO]);
close(recv_pipe[STDOUT_FILENO]);
encoder_pipe[STDIN_FILENO] = recv_pipe[STDIN_FILENO];
encoder_pipe[STDOUT_FILENO] = send_pipe[STDOUT_FILENO];

pthread_create(&child2, &attr, input_encoder_handler, NULL);
pthread_create(&child3, &attr, fifo_output_handler, NULL);
}
return 0;
}

void *input_plugin_handler()
{
char command[COMMAND_SIZE + 1];
//   char send_command[COMMAND_SIZE] = "send";
short *echo_buffer = NULL;
short *boost_buffer = NULL;
char tempstr[MAX_STRING_SIZE];
short *buffer = malloc(sizeof(short)*buffer_rate);
short *b;
int zero = 0;
int size, s, empty = 0, i, x, bs = buffer_size;
int done = 0;

b = buffer;
command[COMMAND_SIZE] = '\0';

while(!done) {
while (((size = read(input_pipe[STDIN_FILENO], command, COMMAND_SIZE)) < 0) && (errno == EAGAIN));

if (echo) 
if (echo_buffer == NULL)
echo_buffer = (short*) calloc(buffer_rate,sizeof(short));
if (boost) 
if (boost_buffer == NULL)
boost_buffer = (short*) calloc(buffer_rate,sizeof(short));
if (size < 0)
{
perror("Player error:");
kill(plugin_pid, SIGKILL);
close(input_pipe[0]);
close(input_pipe[1]);
fork_input_plugin(PLUG_IN);
gdk_threads_enter();
next_track();
gdk_threads_leave();
done = 1;
}
if (strcmp(command, "done") == 0) {
gdk_threads_enter();
if (now_playing)
next_track();
gdk_threads_leave();
} else if (strcmp(command, "options") == 0) {

} else if (strcmp(command, "time") == 0) {
read(input_pipe[STDIN_FILENO], &current_time, sizeof(int));
update_timer(mytimer, current_time);
} else if (strcmp(command, "data") == 0) {
read(input_pipe[STDIN_FILENO], &size, sizeof(int));
while (size)
{
if (size >= bs) {
size -= bs;
while (bs)
{
s = read(input_pipe[STDIN_FILENO], b, bs);
if (s < 0)
{
 if (errno != EAGAIN) {
   perror("Unrecoverable error while reading into data buffer:");
   clean_up_and_quit(1);
   done = 1;
   break;
 }
} else if (s > 0) {
 b = (short*) ((int) b + s);
 bs -= s;
}
}
if (echo) {
for (i = 0; i < buffer_rate; i++)
{
   x = (buffer[i] > 0) ? 1 : -1;
   buffer[i] = echo_buffer[i] = x * (int) sqrt((double) buffer[i]*buffer[i]+echo_buffer[i]);
}
}
if (boost) {
for (i = 0; i < buffer_rate; i++)
{
   for (x = 2; x < 16; x += 2)
   {
       boost_buffer[(i + x) % buffer_rate] = (boost_buffer[(i + x) % buffer_rate] + buffer[i]) / 2;
   }
   x = boost_buffer[i]; // + buffer[i]) / 2;
   buffer[i] = (x > 0) ? -(x-32768)*(x-32768)/32768+32768
		       : (x+32768)*(x+32768)/32768-32768;
}
}
if (linuxout > 0) {
bs = buffer_size;
b = buffer;
while (bs)
{
  s = write(linuxout, b, bs);
  
  if (s == -1)
  {
    if (errno != EINTR) {
	perror("Unrecoverable error writing to sound card");
	break;
    }
  } else if (s > 0) {
    b = (short*) ((int) b + s);
    bs -= s;
  }
}
}
if (shoutcast) {
bs = buffer_size;
b = buffer;
while (bs && !done)
{
  i = write(encoder_pipe[STDOUT_FILENO], b, bs);
  if (i < 0) {
     if (errno != EAGAIN) {
	perror("Unrecoverable error writing to encoder:");
	done = 1;
     }
  } else if (i > 0) {
    b = (short*) ((int) b + i);
    bs -= i;
  }
}
if (done)
  break;
}
if (empty_space) {
for (i = 0; i < buffer_rate; i++)
{
  if (buffer[i] == 0)
     empty++;
  else
     empty = 0;
}
}
bs = buffer_size;
b = buffer;
} else {
while (size)
{
s = read(input_pipe[STDIN_FILENO], b, size);
if (s < 0)
{
 if (errno != EAGAIN) {
    perror("Unrecoverable error while reading into data buffer:");
    clean_up_and_quit(1);
    done = 1;
    break;
 }
} else if (s > 0) {
 b = (short*) ((int) b + s);
 size -= s;
 bs -= s;
}
}
}
}
if (empty_space) {
// empty_space seconds of silence and we go to the next song
if (empty >= freq*empty_space)
{
gdk_threads_enter();
next_track();
gdk_threads_leave();
empty = 0;
}
}
/*       pthread_mutex_lock(&alarm_mutex);
if (++alarm_count >= URATE) {
if (!alarm_pass)
sleep(1);
alarm_pass = 0;
alarm_count = 0;
alarm(1);
}
pthread_mutex_unlock(&alarm_mutex); */
/*       gdk_threads_enter();
if (now_playing && !now_paused) {
write_cmd_handler(send_command);
}
gdk_threads_leave(); */
} else {
printf("Unknown command from plugin (%d): %s\n", size, command);
}
}
pthread_exit(&zero);
}

void* input_encoder_handler()
{
int i, bs, x = 0, done = 0;
int zero = 0;
char *e;

while(!done)
{
e = encode_buffers[x];
bs = buf_size;
// Reads in exactly one second of music.
while ((i = read(encoder_pipe[STDIN_FILENO], e, bs)) < bs)
{
if (i < 0) {
if (errno != EAGAIN) {
perror("Error reading from encoder:");
clean_up_and_quit(1);
done = 1;
}
} else if (i > 0) {
e = (char*) ((int) e + i);
bs -= i;
}
}

pthread_mutex_lock(&shoutcast_mutex);

/* Update current position holder so that new
connections know where to start. */
queue_last_pos = x;

//     printf("queue_last_pos %d\n", x);
x = (x + 1) % MAX_BUFS;

for (i = 0; i < SHOUTCAST_CONNECTIONS_MAX; i++)
{
// If we have a valid queue slot.
if ((map[i] != -1) && (thread_done[i] != 1))
// Kill stalled connection.
// Since x has moved forwards, and there's a lock around
// setting queue_pos[i], then x should never equal queue_pos[i]
// as that'd indicate that queue_pos[i] is MAX_BUFS-1 behind.
if (queue_pos[i] == x) {
printf("Thread %d too far behind.\n", i);
thread_done[i] = 1;
}
}

pthread_cond_broadcast(&shoutcast_cond);

pthread_mutex_unlock(&shoutcast_mutex);
}
pthread_exit(&zero);
}

void* fifo_output_handler()
{
int fifo, pos, done;
char name[MAX_STRING_SIZE];
char* e;
int x, w;

snprintf(name, MAX_STRING_SIZE, "%s/mp3_out", temp_dir);
mkfifo(name, 0);
chmod(name, 0600);
while (1)
{
done = 0;
fifo = open(name, O_WRONLY);
pthread_mutex_lock(&shoutcast_mutex);
pos = queue_last_pos;
pthread_mutex_unlock(&shoutcast_mutex);
while (!done)
{
pthread_mutex_lock(&shoutcast_mutex);

// Wait until there's something to send.
while (pos == queue_last_pos)
pthread_cond_wait(&shoutcast_cond, &shoutcast_mutex);

pthread_mutex_unlock(&shoutcast_mutex);

// Send data.
x = buf_size;
e = encode_buffers[pos];
while (x) {
w = write(fifo, e, x);
if (w < 0) {
if (errno != EAGAIN) {
done = 1;
break;
}
} else {
e += w;
x -= w;
}
}
pos = (pos + 1) % MAX_BUFS;
}
}
}

void* shoutcast_server()
{
int connection, alen;
struct sockaddr_in incoming_sin;
int zero = 0, done = 0;

// Set up listener.
listener = passiveTCP(port, QLEN);

if (listener >= 0)
{
while (!done) {
alen = sizeof(incoming_sin);
connection = accept(listener, (struct sockaddr *)&incoming_sin, &alen);
// Listener is closed which accept doesn't handle too well, so
// we'll just block the thread from doing anything because at
// the point the listener is closed, the mutex lock is activated
// and it's never let go.
pthread_mutex_lock(&clean_up_mutex);
if (listener) {
printf("Connection from %s.\n", inet_ntoa(incoming_sin.sin_addr));
if (connection < 0)
printf("Failed to connect user %s.\n", inet_ntoa(incoming_sin.sin_addr));
else
handle(connection, inet_ntoa(incoming_sin.sin_addr));
} else
done = 1;
pthread_mutex_unlock(&clean_up_mutex);
}
} else
printf("Unable to open port.\n");
pthread_exit(&zero);
}

int handle(int connection, char* name)
{
// A child thread (not too important to store).
int i, w, r;
char buffer[20];
static char* message = "ICY 200 OK\n"
	   "icy-notice1:<BR>This stream requires <a href=\"http://www.xmms.org/\">XMMS</a><BR>\n"
	   "icy-notice2:SHOUTcast Distributed Network Audio Server/Linux/playspc_gkt<BR>\n"
	   "icy-name:playspc_gtk\n"
	   "icy-genre:spc\n"
	   "icy-url:http://www.css.tayloru.edu/~jefoote/playspc/\n"
	   "icy-pub:1\n"
	   "icy-br:";
// A child thread (not too important to store).
pthread_t child;
pthread_attr_t attr;

pthread_attr_init(&attr);
// Make it detached.
pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
// And one-to-one with kernel threads.
pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);

pthread_mutex_lock(&shoutcast_mutex);

if (shoutcast_connections >= SHOUTCAST_CONNECTIONS_MAX) {
pthread_mutex_unlock(&shoutcast_mutex);
close(connection);
return -1;
}

// Read the header (not exactly a bright little reader).
do
{
r = read(connection, buffer, 20);
}
while ((r == 20) || ((r <= 0) && (errno == EAGAIN)));

// Leave on failure
if (r <= 0) {
close(connection);
pthread_mutex_unlock(&shoutcast_mutex);
return -1;
}

// Write out our header.
do
{
w = write(connection, message, strlen(message));
}
while ((w <= 0) && (errno == EAGAIN));

// Leave on failure
if (w <= 0) {
close(connection);
pthread_mutex_unlock(&shoutcast_mutex);
return -1;
}

i = snprintf(buffer, 9, "%d\n\n", bitrate);

// Write out our bitrate.
do
{
w = write(connection, buffer, i);
}
while ((w <= 0) && (errno == EAGAIN));

// Leave on failure
if (w <= 0) {
close(connection);
pthread_mutex_unlock(&shoutcast_mutex);
return -1;
}

for (i = 0; i < SHOUTCAST_CONNECTIONS_MAX; i++)
if (map[i] == -1) {
new_id = i;
map[i] = i;
break;
}

connections = connection;
snprintf(connections_name, 39, "%s", name);

// Fill in the thread's queue information.
queue_pos[new_id] = queue_last_pos;

// Creates a seperate thread to handle input from the spc player.
pthread_create(&child, &attr, shoutcast_client, NULL);

return 0;
}

void* shoutcast_client()
{
int w, x;
int id, one = 1, pos;
int con = connections;
char connection_name[40];
char *e;

id = new_id;
strcpy(connection_name, connections_name);

thread_done[new_id] = 0;

pthread_mutex_unlock(&shoutcast_mutex);

while (!thread_done[id])
{
pthread_mutex_lock(&shoutcast_mutex);

// Wait until there's something to send.
while ((queue_pos[id] == queue_last_pos) && !thread_done[id])
pthread_cond_wait(&shoutcast_cond, &shoutcast_mutex);

if (!thread_done[id]) {

pos = queue_pos[id];
//         printf("%d> queue_pos %d, queue_len %d\n", id, queue_pos[id], queue_len[id]);
queue_pos[id] = (queue_pos[id] + 1) % MAX_BUFS;

pthread_mutex_unlock(&shoutcast_mutex);

// Send data.
x = buf_size;
e = encode_buffers[pos];
while (x) {
w = write(con, e, x);
if (w < 0) {
if (errno != EAGAIN) {
 printf("Thread broke because of unresovable error.\n");
 thread_done[id] = 1;
 break;
}
} else {
e = (char*) ((int) e + w);
x -= w;
}
}
}
}

// And close the connection.
close(con);
printf("Connection from %s lost.\n", connection_name);

pthread_mutex_lock(&shoutcast_mutex);

// Clear the mapping so it can be reused.
map[id] = -1;

shoutcast_connections--;

// Okay to unlock mutex now.
pthread_mutex_unlock(&shoutcast_mutex);
pthread_exit(&one);
}

// Stop playing the current SPC
void stop()
{
const char stop[COMMAND_SIZE] = "stop";
char tempstr[MAX_STRING_SIZE];

//    alarm(0);

if (linuxout > 0)
Shutdown();

if(now_playing) {
now_playing = FALSE;
reset_timer(mytimer);
write_cmd_handler(stop);
}
}

// Pause playing
int pause()
{
const char pause[10] = "pause";
const char unpause[10] = "unpause";

if (now_playing && now_paused) {
g_print("Unpaused playback.\n");
now_paused = FALSE;
write_cmd_handler(unpause);
/*        pthread_mutex_lock(&alarm_mutex);
alarm_count = 0;
alarm_pass = 0;
alarm(1);
pthread_mutex_unlock(&alarm_mutex); */
} else if (now_playing) {
//        alarm(0);
g_print("Paused playback.\n");
now_paused = TRUE;
write_cmd_handler(pause);
}

return 0;
}

/*void alarm_handler(int x)
{
alarm_pass = 1;
} */

// Start playing the currently-selected item in the playlist
int play()
{
gchar* temp;
const char play[10] = "play";
char tempstr[MAX_STRING_SIZE];

if(playlist_length == 0 || current_row == -1) {
g_print("Nothing to play!\n");
return 0;
}

if (linuxout == 0) {
if (OpenLinuxOut()) {
g_print("Error: Unable to open sound device.\n");
return 1;
}
}

// Send file.
write_cmd_handler(play);

if (current_song_length < 0)
   limit = lowerbound;
else {
   if (upperbound) {
      if (current_song_length > upperbound)
         limit = upperbound;
      else
         limit = current_song_length;
   } else {
     limit = current_song_length;
   }
}

total_time = limit;

temp = (gchar*) malloc(sizeof(gchar) * MAX_STRING_SIZE);
gtk_clist_get_text(GTK_CLIST(playlist), current_row, 0, &temp);
g_print("Playing: %s\n", temp);
now_playing = TRUE;
now_paused = FALSE;

define_timer(mytimer, 0, limit);

/*    pthread_mutex_lock(&alarm_mutex);
alarm(1);
alarm_count = 0;
alarm_pass = 0;
pthread_mutex_unlock(&alarm_mutex); */

write_handler(&limit, sizeof(int));
write_handler(playlistentry[current_row].full_filename, MAX_STRING_SIZE);

//    free(temp);
return 0;
}
int OpenLinuxOut(void)
{
int format, stereo, frag;
const char freq_command[COMMAND_SIZE] = "freq";
int freq = FREQ;

linuxout = open("/dev/dsp", O_WRONLY);

if (linuxout == -1)
return 1;

format = AFMT_S16_LE;
stereo = CHANNELS;
frag = 32 << 16;

ioctl(linuxout,SNDCTL_DSP_RESET,0);
ioctl(linuxout,SNDCTL_DSP_POST,0);
//ioctl(linuxout,SNDCTL_DSP_SETFRAGMENT, &frag);
if (ioctl(linuxout,SNDCTL_DSP_SETFMT,&format) == -1)
printf("Error setting format.\n");
if(ioctl(linuxout,SNDCTL_DSP_CHANNELS,&stereo) == -1)
printf("Error selecting mono/stereo.\n");
if (ioctl(linuxout,SNDCTL_DSP_SPEED,&freq) == -1)
printf("Error setting speed.\n");

write_cmd_handler(freq_command);
write_handler(&freq, sizeof(int));

buffer_rate = freq*CHANNELS/URATE;
buffer_size = buffer_rate*sizeof(short);

return 0;
}

void Shutdown()
{
if (linuxout)
close(linuxout);
linuxout = 0;
}

// This is a timeout function that interrupts the GTK main loop at a specified time.
int next_track()
{
if(current_row > -1) {
if (current_row == playlist_length - 1)
   gtk_clist_select_row(GTK_CLIST(playlist), 0, 0);
else
   gtk_clist_select_row(GTK_CLIST(playlist), current_row + 1, 0);
if (now_playing)
   play();
}

return 0;
}

// This is a timeout function that interrupts the GTK main loop at a specified time.
int prev_track()
{
if(current_row > -1) {
if (current_row == 0)
   gtk_clist_select_row(GTK_CLIST(playlist), playlist_length - 1, 0);
else
   gtk_clist_select_row(GTK_CLIST(playlist), current_row - 1, 0);
if (now_playing)
   play();
}

return 0;
}

// Generic callback function
void buttonhandler(GtkWidget* widget, gpointer data)
{
    int temp, tempvar;
    gchar* trash[1];
    char tempstr[MAX_STRING_SIZE];
    int i;

    trash[0] = (gchar*) malloc(sizeof(gchar) * MAX_STRING_SIZE);

    for (i = 0; i < ACOUNT(option_gboolean); i++)
    {
      if (option_gboolean[i].long_name != NULL) {
         if (strncmp((char*) data, option_gboolean[i].long_name, MAX_STRING_SIZE) == 0) {
            if (GTK_TOGGLE_BUTTON(widget)->active)
               temp_gboolean[i] = TRUE;
            else
               temp_gboolean[i] = FALSE;
            return;
         }
      }
    }
       
    for (i = 0; i < ACOUNT(option_int); i++)
    {
      if (option_int[i].long_name != NULL) {
         if (strncmp((char*) data, option_int[i].long_name, MAX_STRING_SIZE) == 0) {
            temp_int[i] = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(widget));
            return;
         }
      }
    }

    for (i = 0; i < ACOUNT(option_string); i++)
    {
      if (option_string[i].long_name != NULL) {
         if (strncmp((char*) data, option_string[i].long_name, MAX_STRING_SIZE) == 0) {
            strncpy(temp_string[i], gtk_entry_get_text(GTK_ENTRY(widget)), MAX_STRING_SIZE);
            return;
         }
      }
    }

    if (strcmp((char*) data, "prefsok") == 0) {
       for (i = 0; i < ACOUNT(option_gboolean); i++)
         *(option_gboolean[i].variable) = temp_gboolean[i];

       for (i = 0; i < ACOUNT(option_int); i++)
         *(option_int[i].variable) = temp_int[i];

       for (i = 0; i < ACOUNT(option_string); i++)
         strncpy(option_string[i].variable, temp_string[i], MAX_STRING_SIZE);

       rebuild_clist();
    } else if(strcmp((char*) data, "prefscancel") == 0) {
       for (i = 0; i < ACOUNT(option_gboolean); i++)
         temp_gboolean[i] = *(option_gboolean[i].variable);

       for (i = 0; i < ACOUNT(option_int); i++)
         temp_int[i] = *(option_int[i].variable);

       for (i = 0; i < ACOUNT(option_string); i++)
         strncpy(temp_string[i], option_string[i].variable, MAX_STRING_SIZE);
    } else if(strcmp((char*) data, "rewind") == 0) {
	prev_track();
    } else if(strcmp((char*) data, "play") == 0) {
	play();
    } else if(strcmp((char*) data, "stop") == 0) {
	stop();
    } else if(strcmp((char*) data, "pause") == 0) {
	pause();
    } else if(strcmp((char*) data, "fastforward") == 0) {
	next_track();
    } else if(strcmp((char*) data, "remove") == 0) {
	if(current_row > -1) {
	    playlist_length--;
	    for(temp = current_row; temp < playlist_length; temp++) {
		strncpy(playlistentry[temp].full_filename, playlistentry[temp + 1].full_filename, MAX_STRING_SIZE);
		strncpy(playlistentry[temp].source_filename, playlistentry[temp + 1].source_filename, MAX_STRING_SIZE);
                playlistentry[temp].fromrar = playlistentry[temp + 1].fromrar;
	    }
	    gtk_clist_remove(GTK_CLIST(playlist), current_row);
	}
    } else if(strcmp((char*) data, "moveup") == 0) {
	if(current_row > 0) {
	    temp = current_row;
	    strncpy(tempstr, playlistentry[current_row].full_filename, MAX_STRING_SIZE);
	    strncpy(playlistentry[current_row].full_filename, playlistentry[current_row - 1].full_filename, MAX_STRING_SIZE);
	    strncpy(playlistentry[current_row - 1].full_filename, tempstr, MAX_STRING_SIZE);
	    strncpy(tempstr, playlistentry[current_row].source_filename, MAX_STRING_SIZE);
	    strncpy(playlistentry[current_row].source_filename, playlistentry[current_row - 1].source_filename, MAX_STRING_SIZE);
	    strncpy(playlistentry[current_row - 1].source_filename, tempstr, MAX_STRING_SIZE);
            tempvar = playlistentry[current_row].fromrar;
            playlistentry[current_row].fromrar = playlistentry[current_row - 1].fromrar;
            playlistentry[current_row - 1].fromrar = tempvar;
	    gtk_clist_get_text(GTK_CLIST(playlist), current_row, 0, &trash[0]);
	    gtk_clist_insert(GTK_CLIST(playlist), current_row - 1, trash);
	    gtk_clist_remove(GTK_CLIST(playlist), current_row + 1);
	    gtk_clist_select_row(GTK_CLIST(playlist), temp - 1, 0);
	}
    } else if(strcmp((char*) data, "movedown") == 0) {
	if (current_row > -1 && current_row < playlist_length) {
	    temp = current_row;
	    strncpy(tempstr, playlistentry[current_row].full_filename, MAX_STRING_SIZE);
	    strncpy(playlistentry[current_row].full_filename, playlistentry[current_row + 1].full_filename, MAX_STRING_SIZE);
	    strncpy(playlistentry[current_row + 1].full_filename, tempstr, MAX_STRING_SIZE);
	    strncpy(tempstr, playlistentry[current_row].source_filename, MAX_STRING_SIZE);
	    strncpy(playlistentry[current_row].source_filename, playlistentry[current_row + 1].source_filename, MAX_STRING_SIZE);
	    strncpy(playlistentry[current_row + 1].source_filename, tempstr, MAX_STRING_SIZE);
            tempvar = playlistentry[current_row].fromrar;
            playlistentry[current_row].fromrar = playlistentry[current_row + 1].fromrar;
            playlistentry[current_row + 1].fromrar = tempvar;
	    gtk_clist_get_text(GTK_CLIST(playlist), current_row, 0, &trash[0]);
	    gtk_clist_insert(GTK_CLIST(playlist), current_row + 2, trash);
	    gtk_clist_remove(GTK_CLIST(playlist), current_row);
	    gtk_clist_select_row(GTK_CLIST(playlist), temp + 1, 0);
	}
    } else if(strcmp((char*) data, "saveid666") == 0) {
	write_id666_tag(playlistentry[current_row].full_filename, &tag);
	fill_name(tempstr,current_row);
        gtk_clist_set_text(GTK_CLIST(playlist), current_row, 0, tempstr);
    } else {
	g_print("Unknown event received: %s\n", (char*) data);
    }
}

void rebuild_clist()
{
    int counter, temp;
    gchar* trash[1];

    temp = current_row;
    current_row = -1;
 
    trash[0] = (gchar*) malloc(sizeof(gchar) * MAX_STRING_SIZE);

    gtk_clist_freeze(GTK_CLIST(playlist));

    gtk_clist_clear(GTK_CLIST(playlist));

    counter = playlist_length;
    for(playlist_length = 0; playlist_length < counter; playlist_length++)
    {
       fill_name(trash[0],playlist_length);
       gtk_clist_append(GTK_CLIST(playlist), trash);
    }

    gtk_clist_thaw(GTK_CLIST(playlist));
    gtk_clist_select_row(GTK_CLIST(playlist), temp, 0);
}

gboolean file_selection_handler_add(GtkWidget* widget, GtkFileSelection* fs)
{
   return file_selection_handler_joined(widget, fs);
}

gboolean file_selection_handler(GtkWidget* widget, GtkFileSelection* fs)
{
   // A file was openned, so we should clear the playlist regardless of how succesful the file load was.
   playlist_length = 0;
   gtk_clist_clear(GTK_CLIST(playlist));
   // Set the current row to an invalid value, then go to "0". This is to select the first item
   // in the playlist without running it (otherwise it's possible that clist_sel sees current_row == row
   // which would cause it to auto-play.
   current_row = -1;
   gtk_clist_select_row(GTK_CLIST(playlist), 0, 0);
   using = USING_NONE;
   return file_selection_handler_joined(widget, fs);
}

// Callback function for the file dialog of the File/Open menu item
gboolean file_selection_handler_joined(GtkWidget* widget, GtkFileSelection* fs)
{
    char tempstr[MAX_STRING_SIZE];
    strncpy(tempstr, gtk_file_selection_get_filename(GTK_FILE_SELECTION(fs)), MAX_STRING_SIZE);
    return open_multi_file_container(tempstr);
}

gboolean open_multi_file_container(char* file)
{

    FILE* INFILE;
    char tempstr[MAX_STRING_SIZE];

    strncpy(old_filename, filename, MAX_STRING_SIZE);
    using_old = using;

    /* The core filename. */
    strncpy(filename, file, MAX_STRING_SIZE);
    gtk_entry_set_text(GTK_ENTRY(filename_textbox), filename);

    strncpy(current_directory, file, MAX_STRING_SIZE);
    strncpy(current_directory, dirname(current_directory), MAX_STRING_SIZE);
    if (chdir(current_directory) != 0)
       perror("Error changing to directory specified");
    if((strncasecmp(&filename[strlen(filename)-4], ".rar", 4) == 0) ||
       (strncasecmp(&filename[strlen(filename)-4], ".rsn", 4) == 0)) {
	return rar_extract(filename);
    }

    using = USING_PLAYLIST;

    INFILE = fopen(file, "r");
    if(INFILE == NULL) {
	g_print("1> Cannot open %s for reading.  Giving up.\n", filename);
	return FALSE;
    } else {
	if (fgets(tempstr, MAX_STRING_SIZE, INFILE) != NULL) {
	    if (strcmp(tempstr, "PLAYSPC PLAYLIST FILE\n") == 0)
               load_playlist_v1(INFILE);
	    else if (strcmp(tempstr, "PLAYSPC PLAYLIST FILE V2\n") == 0)
               load_playlist_v2(INFILE);
	    else {
                snprintf(tempstr, MAX_STRING_SIZE, "Error, %s is not a valid playlist file!\n", filename);
                g_print(tempstr);
                message_box("Error!", tempstr);
                strncpy(filename, old_filename, MAX_STRING_SIZE);
                gtk_entry_set_text(GTK_ENTRY(filename_textbox), filename);
                using = using_old;
	    }
	} else {
          snprintf(tempstr, MAX_STRING_SIZE, "Error, %s is not a valid playlist file!\n", filename);
          g_print(tempstr);
          message_box("Error!", tempstr);
          strncpy(filename, old_filename, MAX_STRING_SIZE);
          gtk_entry_set_text(GTK_ENTRY(filename_textbox), filename);
          using = using_old;
	}

	fclose(INFILE);
    }

    return TRUE;
}

void load_playlist_v1(FILE* INFILE)
{
    char tempstr[MAX_STRING_SIZE];
    int i;
    int playlist_length_start = playlist_length;
    gchar* trash[1];

    trash[0] = (gchar*) malloc(sizeof(gchar) * MAX_STRING_SIZE);

    while(fgets(tempstr, MAX_STRING_SIZE, INFILE) != NULL) {
       // Strip off the newline that fgets might slap on the end of our string
       for(i = 0; i < MAX_STRING_SIZE; i++) {
          if (tempstr[i] == '\n') {
             tempstr[i] = '\0';
             break;
 	  }
       }
       strncpy(playlistentry[playlist_length].full_filename, tempstr, MAX_STRING_SIZE);
       strncpy(playlistentry[playlist_length].source_filename, filename, MAX_STRING_SIZE);
       playlistentry[playlist_length].fromrar = 0;
       fill_name(trash[0],playlist_length);
       gtk_clist_append(GTK_CLIST(playlist), trash);
       playlist_length++;
   }
   if (playlist_length_start == playlist_length) {
      snprintf(tempstr, MAX_STRING_SIZE, "Error, %s appears to be an empty playlist file!\n", filename);
      g_print(tempstr);
      message_box("Error!", tempstr);
   }
//   free(trash[0]);
}

void load_playlist_v2(FILE* INFILE)
{
    char tempstr[MAX_STRING_SIZE];
    char *name = NULL;
    int i;
    int playlist_length_start = playlist_length;
    gchar* trash[1];

    trash[0] = (gchar*) malloc(sizeof(gchar) * MAX_STRING_SIZE);

    while(fgets(tempstr, MAX_STRING_SIZE, INFILE) != NULL) {
       // Strip off the newline that fgets might slap on the end of our string
       for(i = 0; i < MAX_STRING_SIZE; i++) {
          if (tempstr[i] == '\n') {
             tempstr[i] = '\0';
             break;
 	  }
       }
       for(i = 0; i < MAX_STRING_SIZE; i++) {
          if (tempstr[i] == ',') {
             tempstr[i] = '\0';
             if (i == MAX_STRING_SIZE - 1) {
                snprintf(tempstr, MAX_STRING_SIZE, "Error, %s appears to be corrupt!\n", filename);
                g_print(tempstr);
                message_box("Error!", tempstr);
                playlist_length = playlist_length_start;
                strncpy(filename, old_filename, MAX_STRING_SIZE);
                gtk_entry_set_text(GTK_ENTRY(filename_textbox), filename);
                using = using_old;
                return;
             } else
                name = &tempstr[i+1];
             break;
 	  }
       }

       if (name == NULL) {
          snprintf(tempstr, MAX_STRING_SIZE, "Error, %s appears to be corrupt!\n", filename);
          g_print(tempstr);
          message_box("Error!", tempstr);
          playlist_length = playlist_length_start;
          strncpy(filename, old_filename, MAX_STRING_SIZE);
          gtk_entry_set_text(GTK_ENTRY(filename_textbox), filename);
          using = using_old;
//          free(trash[0]);
          return;
       }

       if (tempstr[0] == '\0') {
          strncpy(playlistentry[playlist_length].full_filename, name, MAX_STRING_SIZE);
          strncpy(playlistentry[playlist_length].source_filename, filename, MAX_STRING_SIZE);
          playlistentry[playlist_length].fromrar = 0;
       } else {
          strncpy(playlistentry[playlist_length].source_filename, tempstr, MAX_STRING_SIZE);
          if (!load_from_rar(name, tempstr)) {
             snprintf(tempstr, MAX_STRING_SIZE, "Error, %s wasn't extracted properly from %s\n", name, tempstr);
             g_print(tempstr);
             message_box("Error!", tempstr);
             playlist_length = playlist_length_start;
             strncpy(filename, old_filename, MAX_STRING_SIZE);
             gtk_entry_set_text(GTK_ENTRY(filename_textbox), filename);
             using = using_old;
//             free(trash[0]);
             return;
          }
          strncpy(playlistentry[playlist_length].full_filename, tempstr, MAX_STRING_SIZE);
          playlistentry[playlist_length].fromrar = 1;
       }
       fill_name(trash[0],playlist_length);
       gtk_clist_append(GTK_CLIST(playlist), trash);
       playlist_length++;
   }
   if (playlist_length_start == playlist_length) {
      snprintf(tempstr, MAX_STRING_SIZE, "Error, %s appears to be an empty playlist file!\n", filename);
      g_print(tempstr);
      message_box("Error!", tempstr);
      strncpy(filename, old_filename, MAX_STRING_SIZE);
      gtk_entry_set_text(GTK_ENTRY(filename_textbox), filename);
      using = using_old;
   }
//   free(trash[0]);
}

int load_from_rar(char* name, char* file)
{
    FILE* INFILE;
    char temp[MAX_STRING_SIZE];
    char cmdstr[MAX_STRING_SIZE];

    // Extract file.
    snprintf(cmdstr, MAX_STRING_SIZE, "%s%s \"%s\" \"%s\" %s/",
                                      RAR, RAR_EXTRACT_FLAGS, file, name,
				      temp_dir);
    system(cmdstr);

    snprintf(temp, MAX_STRING_SIZE, "%s/%s", temp_dir, name);

    INFILE = fopen(temp, "r");
    if (INFILE == NULL) {
       return 0;
    }

    fclose(INFILE);

    strncpy(file, temp, MAX_STRING_SIZE);

    return 1;
}

// Callback function for the file dialog of the File/Save As menu item
void file_save_playlist_handler(GtkWidget* widget, GtkFileSelection* fs)
{
    snprintf(filename, MAX_STRING_SIZE, "%s", gtk_file_selection_get_filename(GTK_FILE_SELECTION(fs)));
    gtk_entry_set_text(GTK_ENTRY(filename_textbox), filename);
    save_playlist();
}

// Callback function for the file dialog of the File/Save As menu item
void file_save_rar_handler(GtkWidget* widget, GtkFileSelection* fs)
{
    snprintf(filename, MAX_STRING_SIZE, "%s", gtk_file_selection_get_filename(GTK_FILE_SELECTION(fs)));
    gtk_entry_set_text(GTK_ENTRY(filename_textbox), filename);
    save_rar();
}

// Callback function for the file dialog of the ADD toolbar button
void file_add_handler(GtkWidget* widget, GtkFileSelection* fs)
{
    gchar* trash[1];

    trash[0] = (gchar*) malloc(sizeof(gchar) * MAX_STRING_SIZE);

    strcpy(playlistentry[playlist_length].full_filename, gtk_file_selection_get_filename(GTK_FILE_SELECTION(fs)));
    g_print(playlistentry[playlist_length].full_filename);
    g_print("\n");
    strcpy(playlistentry[playlist_length].source_filename, "<Unknown>");
    playlistentry[playlist_length].fromrar = 0;
    fill_name(trash[0],playlist_length);
    gtk_clist_append(GTK_CLIST(playlist), trash);
    playlist_length++;
//    free(trash[0]);
}

// Signal handler for changing text.
void cb_text_entry(GtkWidget* widget, gpointer data)
{
    strncpy(((struct char_and_int*) data)->string, gtk_entry_get_text(GTK_ENTRY(widget)), ((struct char_and_int*) data)->integer);
}

// Signal handler for changing the song length from the 'View File Info' dialog
void cb_id666_songlength(GtkWidget* widget, gpointer data)
{
    gchar* entry_text = gtk_entry_get_text(GTK_ENTRY(widget));

    tag.songlength_raw = atoi(entry_text);

    snprintf(tag.songlength, 6, "%d:%02d", tag.songlength_raw / 60,
                                           tag.songlength_raw % 60);
}

// Signal handler for selecting an item from the clist (playlist)
void clist_sel(GtkWidget* clist, gint row, gint col, GdkEventButton* event, gpointer data)
{
    gchar* text;
    struct id666 tag;

    text = (gchar*) malloc(sizeof(gchar) * MAX_STRING_SIZE);
    gtk_clist_get_text(GTK_CLIST(clist), row, col, &text);
    // If already selected, then they must be wanting to play it.  Interestingly,
    // GTK automatically handles this like a double click case. Ie, if you click
    // too slowly, it will see it as a select and then an unselect.
    if (current_row == row) {
      play();
    } else {
      current_row = row;
      if(true_option_id666 && true_option_obey_length && read_id666_tag(playlistentry[row].full_filename, &tag)) {
        current_song_length = tag.songlength_raw;
      }
      else {
  	current_song_length = -1;
      }
    }
//    free(text);
}

// Signal handler for unselecting an item from the clist (playlist)
void clist_unsel(GtkWidget* clist, gint row, gint col, GdkEventButton* event, gpointer data)
{
    current_row = -1;
    current_song_length = -1;
}

// Signal handler for closing the Preferences or About dialogs
gint destroy(GtkWidget* widget, GdkEvent* event, gpointer data)
{
    return(FALSE);
}

// Callback function for the File/New menu item
void cb_new(GtkWidget* widget, gpointer data)
{
    gtk_clist_clear(GTK_CLIST(playlist));
    current_row = -1;
    filename[0] = '\0';
}

// Callback function for the File/Open menu item
void cb_open(GtkWidget* widget, gpointer data)
{
    GtkWidget* filedialog;

    filedialog = gtk_file_selection_new("Open a playlist file");
    gtk_signal_connect(GTK_OBJECT(filedialog), "destroy", GTK_SIGNAL_FUNC(destroy), "WM destroy");
    gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filedialog)->ok_button), "clicked", GTK_SIGNAL_FUNC(file_selection_handler), filedialog);
    gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(filedialog)->ok_button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(filedialog));
    gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(filedialog)->cancel_button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(filedialog));
    //gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(filedialog)->help_button), "clicked", GTK_SIGNAL_FUNC(cb_open_help), GTK_OBJECT(filedialog));
    gtk_widget_show(filedialog);
    // Don't clear clist hastefully.
}

void cb_open_add(GtkWidget* widget, gpointer data)
{
    GtkWidget* filedialog;

    filedialog = gtk_file_selection_new("Open a playlist file");
    gtk_signal_connect(GTK_OBJECT(filedialog), "destroy", GTK_SIGNAL_FUNC(destroy), "WM destroy");
    gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filedialog)->ok_button), "clicked", GTK_SIGNAL_FUNC(file_selection_handler_add), filedialog);
    gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(filedialog)->ok_button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(filedialog));
    gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(filedialog)->cancel_button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(filedialog));
    //gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(filedialog)->help_button), "clicked", GTK_SIGNAL_FUNC(cb_open_help), GTK_OBJECT(filedialog));
    gtk_widget_show(filedialog);
    // Don't clear clist hastefully.
}

// Callback function for the File/Save menu item
void cb_save(gpointer data, guint action, GtkWidget* widget)
{
    GtkWidget* filedialog;

    if (action != 0)
    {

        if (playlist_length == 0) {
           message_box("Playspc_gtk", "Nothing to save.");
           return;
        }

        if (action == 1) {
           filedialog = gtk_file_selection_new("Save playlist as");
           gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filedialog)->ok_button), "clicked",
                              GTK_SIGNAL_FUNC(file_save_playlist_handler), filedialog);
        } else {
           filedialog = gtk_file_selection_new("Save RAR as");
           gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filedialog)->ok_button), "clicked",
                              GTK_SIGNAL_FUNC(file_save_rar_handler), filedialog);
        }
	gtk_signal_connect(GTK_OBJECT(filedialog), "destroy", GTK_SIGNAL_FUNC(destroy), "WM destroy");
	gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(filedialog)->ok_button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(filedialog));
	gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(filedialog)->cancel_button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(filedialog));
	//gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(filedialog)->help_button), "clicked", GTK_SIGNAL_FUNC(cb_save_help), GTK_OBJECT(filedialog));
	gtk_widget_show(filedialog);
    } else if (filename[0] == '\0') {
        if (playlist_length == 0) {
           message_box("Playspc_gtk", "You have not specified a filename. Input one, or use one of the other Save As'.");
        }
    } else {

        if ((using == USING_NONE) || (using == USING_CUSTOM)) {
           message_box("Playspc_gtk", "You must specify how to save.");
        } else if (playlist_length == 0) {
           message_box("Playspc_gtk", "Nothing to save.");
        } else if (using == USING_RAR) {
           save_rar();
        } else if (using == USING_PLAYLIST) {
           save_playlist();
        } else
           message_box("Error!", "Uh oh. This shouldn't ever happen.\n");
    }
}

void save_rar()
{

// A child thread (not too important to store).
pthread_t child;
pthread_attr_t attr;

// Locks the clean_up_mutex so we can't delete the temporaries until
// *after* the rar is created.

pthread_mutex_lock(&clean_up_mutex);

// Unless, of course, save_rar was flat out called too late so we should
// abort.

if (remove_temps_done) {
   g_print("Too late to save.  Temporaries have already been deleted.\n");
   pthread_mutex_unlock(&clean_up_mutex);
   return;
}

// This is threaded so the saving doesn't lock up the GUI.

pthread_attr_init(&attr);
// Make it detached.
pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
// And one-to-one with kernel threads.
pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);

// Creates a seperate thread to handle input from the spc player.
pthread_create(&child, &attr, saving_rar, NULL);

}

void* saving_rar()
{
    int zero = 0;
    char cmdstr[65536];
    char tmp[65536];
    int counter;
    FILE* file;
    char name[MAX_STRING_SIZE];

    snprintf(name, MAX_STRING_SIZE, "%s/playlist", temp_dir);

    file = fopen(name, "w");

    for (counter = 0; counter < playlist_length; counter++) {
        if (playlistentry[counter].fromrar) {
           // drop the tempdir part
           fputs(&playlistentry[counter].full_filename[strlen(temp_dir) + 1], file);
        } else {
           // drop the preceding slash
           fputs(&playlistentry[counter].full_filename[1], file);
        }
	if(counter < playlist_length - 1)
	    fputs("\n", file);
    }

    fclose(file);


    chdir(temp_dir);

    snprintf(cmdstr, 65536, RAR RAR_COMPRESS_FLAGS " \"%s\" playlist", filename);

    for(counter = 0; counter < playlist_length; counter++)
    {
       strncpy(tmp, cmdstr, 65536);
       if (playlistentry[counter].fromrar)
         snprintf(cmdstr, 65536, "%s \"%s\"", tmp, &playlistentry[counter].full_filename[strlen(temp_dir) + 1]);
       else
         snprintf(cmdstr, 65536, "%s \"%s\"", tmp, playlistentry[counter].full_filename);
    }

    strncpy(tmp, cmdstr, 65536);
    snprintf(cmdstr, 65536, "%s > /dev/null", tmp);

    system(cmdstr);
    
    pthread_mutex_unlock(&clean_up_mutex);
    pthread_exit(&zero);
}


void save_playlist()
{
    char error[MAX_STRING_SIZE];
    int counter;
    FILE* OUTFILE;

    OUTFILE = fopen(filename, "w");
    if (OUTFILE == NULL) {
       snprintf(error, MAX_STRING_SIZE, "Cannot open %s for writing.  Giving up.\n", filename);
       g_print(error);
       message_box("Error!", error);
    } else {
      fputs("PLAYSPC PLAYLIST FILE V2\n", OUTFILE);
      for(counter = 0; counter < playlist_length; counter++) {
	if (playlistentry[counter].fromrar) {
           // drop the tempdir part
           fputs(playlistentry[counter].source_filename, OUTFILE);
           fputs(",", OUTFILE);
           fputs(&playlistentry[counter].full_filename[strlen(temp_dir) + 1], OUTFILE);
        } else {
           fputs(",", OUTFILE);
           fputs(playlistentry[counter].full_filename, OUTFILE);
        }
	if(counter < playlist_length - 1)
	    fputs("\n", OUTFILE);
      }
      fclose(OUTFILE);
    }
}

void message_box(char* title, char* text)
{
    GtkWidget* win;
    GtkWidget* button;
    GtkWidget* label;
    GtkWidget* table;

    win = gtk_window_new(GTK_WINDOW_DIALOG);
    gtk_window_set_default_size(GTK_WINDOW(win), 300, 75);
    gtk_window_set_title(GTK_WINDOW(win), title);
    gtk_container_set_border_width(GTK_CONTAINER(win), 10);
    gtk_signal_connect(GTK_OBJECT(win), "destroy", GTK_SIGNAL_FUNC(destroy), "WM destroy");

    table = gtk_vbox_new(FALSE, 0);
    gtk_container_add(GTK_CONTAINER(win), table);
    gtk_widget_show(table);

    label = gtk_label_new(text);
    gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
    gtk_box_pack_start(GTK_BOX(table), label, FALSE, TRUE, 0);
    gtk_widget_show(label);

    button = gtk_button_new_with_label("Okay");
    gtk_signal_connect_object(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(win));
    gtk_box_pack_start(GTK_BOX(table), button, FALSE, TRUE, 0);
    gtk_widget_show(button);

    gtk_widget_show(win);
}


// Callback for the ADD toolbar button
void cb_add(GtkWidget* widget, gpointer data)
{
    GtkWidget* filedialog;

    filedialog = gtk_file_selection_new("Add to playlist");
    gtk_signal_connect(GTK_OBJECT(filedialog), "destroy", GTK_SIGNAL_FUNC(destroy), "WM destroy");
    gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(filedialog)->ok_button), "clicked", GTK_SIGNAL_FUNC(file_add_handler), filedialog);
    gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(filedialog)->ok_button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(filedialog));
    gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(filedialog)->cancel_button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(filedialog));
    //gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(filedialog)->help_button), "clicked", GTK_SIGNAL_FUNC(cb_add_help), GTK_OBJECT(filedialog));
    gtk_widget_show(filedialog);
}

GtkWidget* gtk_label_and_text_box(GtkWidget* parent, char* label_name, int max_len)
{
    GtkWidget *box, *label, *textfield;

    box = gtk_hbox_new(FALSE, 0);
    label = gtk_label_new(label_name);
    gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_RIGHT);
    gtk_label_set_line_wrap(GTK_LABEL(label), FALSE);
    textfield = gtk_entry_new_with_max_length(max_len);
    gtk_box_pack_start(GTK_BOX(box), textfield, TRUE, TRUE, 0);
    gtk_box_pack_start(GTK_BOX(box), label, TRUE, TRUE, 0);
    gtk_box_pack_start(GTK_BOX(parent), box, TRUE, TRUE, 0);
    gtk_widget_show(label);
    gtk_widget_show(textfield);
    gtk_widget_show(box);

    return textfield;
}

// Callback function for the Options/Preferences menu item
void cb_prefs(GtkWidget* widget, gpointer data)
{
    GtkWidget* prefswin;
    GtkWidget* button;
    GtkWidget* buttons;
    GtkWidget* table;
    GtkWidget* checkbox;
    GtkWidget* textfield;
    GtkWidget* spinbox;
    GtkTooltips* tooltips;
    int i;

    tooltips = gtk_tooltips_new();

    prefswin = gtk_window_new(GTK_WINDOW_DIALOG);
    gtk_window_set_title(GTK_WINDOW(prefswin), "Preferences");
    gtk_container_set_border_width(GTK_CONTAINER(prefswin), 0);
    gtk_signal_connect(GTK_OBJECT(prefswin), "destroy", GTK_SIGNAL_FUNC(destroy), "WM destroy");

    table = gtk_vbox_new(TRUE, 0);
    gtk_container_add(GTK_CONTAINER(prefswin), table);
    gtk_widget_show(table);

    for (i = 0; i < ACOUNT(option_gboolean); i++)
    {
      if (option_gboolean[i].long_name != NULL) {
         checkbox = gtk_check_button_new_with_label(option_gboolean[i].long_name);
         gtk_tooltips_set_tip(tooltips, checkbox, option_gboolean[i].description, NULL);
         if (*(option_gboolean[i].variable)) {
            gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbox), TRUE);
         }
         gtk_signal_connect(GTK_OBJECT(checkbox), "toggled", GTK_SIGNAL_FUNC(buttonhandler), (gpointer) option_gboolean[i].long_name);
         gtk_box_pack_start(GTK_BOX(table), checkbox, TRUE, TRUE, 0);
         gtk_widget_show(checkbox);
      }
    }

    for (i = 0; i < ACOUNT(option_string); i++)
    {
      if (option_string[i].long_name != NULL) {
         textfield = gtk_label_and_text_box(table, (char*) option_string[i].long_name, option_string[i].max_length);
         gtk_tooltips_set_tip(tooltips, textfield, option_string[i].description, NULL);
         gtk_entry_set_text(GTK_ENTRY(textfield), option_string[i].variable);
         gtk_signal_connect(GTK_OBJECT(textfield), "changed", GTK_SIGNAL_FUNC(buttonhandler), (gpointer) option_string[i].long_name);
      }
    }

    for (i = 0; i < ACOUNT(option_int); i++)
    {
      if (option_int[i].long_name != NULL) {
         spinbox = gtk_spinbutton_new_with_label(table, (char*) option_int[i].long_name, *(option_int[i].variable),
                                                 option_int[i].min, option_int[i].max, 1, 5, .01);
         gtk_tooltips_set_tip(tooltips, spinbox, option_int[i].description, NULL);
         gtk_signal_connect(GTK_OBJECT(spinbox), "changed", GTK_SIGNAL_FUNC(buttonhandler), (gpointer) option_int[i].long_name);
      }
    }

    buttons = gtk_hbox_new(TRUE, 0);
    button = gtk_button_new_with_label("OK");
    gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(buttonhandler), (gpointer) "prefsok");
    gtk_signal_connect_object(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(prefswin));
    gtk_box_pack_start(GTK_BOX(buttons), button, FALSE, TRUE, 0);
    gtk_widget_show(button);

    button = gtk_button_new_with_label("Cancel");
    gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(buttonhandler), (gpointer) "prefscancel");
    gtk_signal_connect_object(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(prefswin));
    gtk_box_pack_start(GTK_BOX(buttons), button, FALSE, TRUE, 0);
    gtk_widget_show(button);

    gtk_box_pack_end(GTK_BOX(table), buttons, TRUE, TRUE, 0);
    gtk_widget_show(buttons);

    gtk_widget_show(prefswin);
}

GtkWidget* gtk_spinbutton_new_with_label(GtkWidget* table, char* text,
                                         gfloat value, gfloat lower, gfloat upper, gfloat step_increment, gfloat page_increment,
                                         gfloat climb_rate)
{
   GtkWidget* box, *label, *spinbutton;
   GtkObject* adj;

   box = gtk_hbox_new(FALSE, 0);
   adj = gtk_adjustment_new(value, lower, upper, step_increment, page_increment, 0);
   label = gtk_label_new(text);
   spinbutton = gtk_spin_button_new(GTK_ADJUSTMENT(adj), climb_rate, 0);
   gtk_widget_show(label);
   gtk_widget_show(spinbutton);
   gtk_box_pack_start(GTK_BOX(box), spinbutton, TRUE, TRUE, 0);
   gtk_box_pack_end(GTK_BOX(box), label, FALSE, FALSE, 0);
   gtk_widget_show(box);
   gtk_box_pack_start(GTK_BOX(table), box, FALSE, TRUE, 0);

   return spinbutton;   
}

// Callback function for the Options/View File Info menu item
void cb_view_info(GtkWidget* widget, gpointer data)
{
    GtkWidget* infowin;
    GtkWidget* button;
    GtkWidget* buttons;
    GtkWidget* textfield;
    GtkWidget* table;
    char tempstr[MAX_STRING_SIZE];

    if (current_row > -1) {
      read_id666_tag(playlistentry[current_row].full_filename, &tag);
    
      infowin = gtk_window_new(GTK_WINDOW_DIALOG);
      gtk_window_set_title(GTK_WINDOW(infowin), "View File Info");
      gtk_container_set_border_width(GTK_CONTAINER(infowin), 10);
      gtk_signal_connect(GTK_OBJECT(infowin), "destroy", GTK_SIGNAL_FUNC(destroy), "WM destroy");

      table = gtk_vbox_new(FALSE, 0);
      gtk_container_add(GTK_CONTAINER(infowin), table);
      gtk_widget_show(table);

      gtk_label_and_textfield(table,textfield,"Game Title: ",cb_text_entry,tag_gametitle);
      gtk_label_and_textfield(table,textfield,"Song Title: ",cb_text_entry,tag_songtitle);

      tag_songlength.string = tempstr;
      snprintf(tempstr, 4, "%d", tag.songlength_raw); // itoa()
      gtk_label_and_textfield(table,textfield,"Song Length: ",cb_id666_songlength,tag_songlength);

      gtk_label_and_textfield(table,textfield,"Song Author: ",cb_text_entry,tag_songauthor);
      gtk_label_and_textfield(table,textfield,"Dumped by: ",cb_text_entry,tag_dumpername);
      gtk_label_and_textfield(table,textfield,"Dumped on: ",cb_text_entry,tag_dumpdate);
      gtk_label_and_textfield(table,textfield,"Dumped using: ",cb_text_entry,tag_dumpprog);
      gtk_label_and_textfield(table,textfield,"Dumper's Comments: ",cb_text_entry,tag_comments);

      buttons = gtk_hbox_new(TRUE, 0);
      button = gtk_button_new_with_label("Save");
      gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(buttonhandler), (gpointer) "saveid666");
      gtk_signal_connect_object(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(infowin));
      gtk_box_pack_start(GTK_BOX(buttons), button, FALSE, TRUE, 0);
      gtk_widget_show(button);

      button = gtk_button_new_with_label("Cancel");
      gtk_signal_connect_object(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(infowin));
      gtk_box_pack_start(GTK_BOX(buttons), button, FALSE, TRUE, 0);
      gtk_widget_show(button);

      gtk_box_pack_start(GTK_BOX(table), buttons, FALSE, TRUE, 0);
      gtk_widget_show(buttons);
      
      gtk_widget_show(infowin);
   } else
     message_box("Error!", "You must select something first to view its info.");
}

// Callback function for the Help/About menu item
void cb_about(GtkWidget* widget, gpointer data)
{
    GtkWidget* aboutwin;
    GtkWidget* button;
    GtkWidget* label;
    GtkWidget* table;

    aboutwin = gtk_window_new(GTK_WINDOW_DIALOG);
    gtk_window_set_title(GTK_WINDOW(aboutwin), "About");
    gtk_container_set_border_width(GTK_CONTAINER(aboutwin), 10);
    gtk_signal_connect(GTK_OBJECT(aboutwin), "destroy", GTK_SIGNAL_FUNC(destroy), "WM destroy");

    table = gtk_vbox_new(FALSE, 0);
    gtk_container_add(GTK_CONTAINER(aboutwin), table);
    gtk_widget_show(table);

    label = gtk_label_new(
       "PlaySPC_GTK\nversion " CURRENT_VERSION "\n"
       "Copyright (C) 2002, JJ Foote\n\n"
       "This program is free software; you can redistribute it and/or modify\n" 
       "it under the terms of the GNU General Public License as published by\n"
       "the Free Software Foundation; either version 2 of the License, or\n"
       "(at your option) any later version.\n\n"
       "This program is distributed in the hope that it will be useful,\n"
       "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
       "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"
       "See the GNU General Public License for more details.\n");
    gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);
    gtk_label_set_line_wrap(GTK_LABEL(label), FALSE);
    gtk_label_set_pattern(GTK_LABEL(label), "___________");
    gtk_box_pack_start(GTK_BOX(table), label, FALSE, TRUE, 0);
    gtk_widget_show(label);

    button = gtk_button_new_with_label("Dismiss");
    gtk_signal_connect_object(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(aboutwin));
    gtk_box_pack_start(GTK_BOX(table), button, FALSE, TRUE, 0);
    gtk_widget_show(button);

    gtk_widget_show(aboutwin);
}

// Clean up properly when we quit the application
void clean_up_and_quit(int x)
{
    char tempstr[MAX_STRING_SIZE];
    int i;
//    FILE* INFILE;

    // Only allow one thread to clean up.
    pthread_mutex_lock(&clean_up_mutex);

    if (!save_rc_done)
       save_rc(); // Save preferences to ~/.playspc_gtkrc file
    save_rc_done = 1;
    if (plugin_pid)
       kill(plugin_pid, SIGTERM);
    plugin_pid = 0;
    if (shoutcast) {
       if (listener)
         close(listener);
       listener = 0;
       pthread_mutex_lock(&shoutcast_mutex);
       if (encoder_pid)
         kill(encoder_pid, SIGTERM);
       for (i = 0; i < SHOUTCAST_CONNECTIONS_MAX; i++)
         thread_done[i] = 1;
       unlink("/tmp/mp3_out");
       pthread_mutex_unlock(&shoutcast_mutex);
       pthread_cond_broadcast(&shoutcast_cond);
    }
    shoutcast = 0;

    if (!remove_temps_done) {
      // Remove temporary directory and all that it contains.
      strcpy(tempstr, RM);
      strcat(tempstr, RM_FLAGS);
      strcat(tempstr, temp_dir);
      system(tempstr);
    }

    remove_temps_done = 1;
    
    if (!gtk_done) {
      gtk_done = 1;
      pthread_mutex_unlock(&clean_up_mutex);
      // We're done; tell GTK that it can quit now
      gtk_main_quit();
    }
      
    pthread_mutex_unlock(&clean_up_mutex);
}

void input_quit()
{
   const char exit[10] = "exit";

   write_cmd_handler(exit);
}

// Set up the menubar
static GtkItemFactoryEntry menu_items[] = {
    { "/_File",			NULL,		NULL,		    0,  "<Branch>"},
    {"/File/_New",		"<control>N",   cb_new,		    0,  NULL},
    {"/File/_Open",		"<control>O",   cb_open,	    0,  NULL},
    {"/File/Open_Add",		"<control>A",   cb_open_add,	    0,  NULL},
    {"/File/_Save As Last",	"<control>S",   cb_save,	    0,  NULL},
    {"/File/Save As _Playlist",	"<control>P",	cb_save,	    1,  NULL},
    {"/File/Save As _Rar",	"<control>R",	cb_save,	    2,  NULL},
    {"/File/sep1",		NULL,		NULL,		    0,  "<Separator>"},
    {"/File/E_xit",		"<control>Q",   clean_up_and_quit,  0,  NULL},
    {"/_Options",		NULL,		NULL,		    0,  "<Branch>"},
    {"/Options/_View File Info",NULL,		cb_view_info,	    0,	NULL},
    {"/Options/_Preferences",	NULL,		cb_prefs,	    0,  NULL},
    {"/_Help",			NULL,		NULL,		    0,  "<LastBranch>"},
    {"/Help/About",		NULL,		cb_about,	    0,  NULL},
};

// Generate the menus
void get_main_menu(GtkWidget* window, GtkWidget** menubar)
{
    GtkItemFactory* itemfactory;
    GtkAccelGroup* accelgroup;
    gint nmenu_items = sizeof(menu_items)/sizeof(menu_items[0]);

    accelgroup = gtk_accel_group_new();
    itemfactory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<main>", accelgroup);
    gtk_item_factory_create_items(itemfactory, nmenu_items, menu_items, NULL);
    gtk_window_add_accel_group(GTK_WINDOW(window), accelgroup);

    if(menubar)
	*menubar = gtk_item_factory_get_widget(itemfactory, "<main>");
}

/******************************
 *            MAIN            *
 ******************************/
int main(int argc, char* argv[])
{
    GtkWidget* window, // Reusable GTK widget objects
	     * vertical_box,
	     * horizontal_toolbar_box,
	     * menubar,
	     * toolbar,
             * filename_label,
             * filename_box,
	     * button,
	     * playlistwin,
	     * iconwidget;
    GdkPixmap* addicon, // xpm images used on toolbar buttons
	     * removeicon,
	     * moveupicon,
	     * movedownicon,
	     * playicon,
	     * pauseicon,
	     * stopicon,
	     * fastforwardicon,
	     * rewindicon;
    GdkBitmap* mask; // transparency mask for pixmaps
    int i, autoplay = 0;

    if (mkdir(TEMP_DIR, 0)) {
       if (errno != EEXIST) {
          g_print("Unable to generate base temporary directory.\n");
          return 1;
       }
    }

    chmod(TEMP_DIR, 0700);

    // Generate a temporary directory using template temp_dir.
    if (mkdtemp(temp_dir) == NULL) {
       g_print("Unable to generate temporary directory.\n");
       return 1;
    }

    // Get current directory, as in worst case we need to know where we start from from the rc file.
    if (getcwd(current_directory, MAX_STRING_SIZE) == NULL)
       g_print("Couldn't get current directory.\n");

    if (init() != 0)
    {
       g_print("Error in initialization.\n");
       return 1;
    }
    
    g_thread_init(NULL);

    gtk_init(&argc, &argv); // Initialize GTK

    // Build the primary window for the application
    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(window), "PlaySPC_GTK");
    gtk_widget_set_usize(GTK_WIDGET(window), 450, 400);
    gtk_signal_connect(GTK_OBJECT(window), "destroy", GTK_SIGNAL_FUNC(clean_up_and_quit), "WM destroy");
    gtk_container_set_border_width(GTK_CONTAINER(window), 0);
    gtk_widget_realize(window); // need this for xpm images to show up properly

    // Build a table inside the window to organize everything nicely
    vertical_box = gtk_vbox_new(FALSE, 0);
    gtk_container_add(GTK_CONTAINER(window), vertical_box);

    // Add the menubar
    get_main_menu(window, &menubar);
    gtk_box_pack_start(GTK_BOX(vertical_box), menubar, FALSE, TRUE, 0);
    gtk_widget_show(menubar);

    horizontal_toolbar_box = gtk_hbox_new(TRUE, 0);
    
    // Build toolbar for basic playback functions
    toolbar = gtk_toolbar_new(GTK_ORIENTATION_HORIZONTAL, GTK_TOOLBAR_ICONS);
    gtk_container_set_border_width(GTK_CONTAINER(toolbar), 5);

    rewindicon = gdk_pixmap_create_from_xpm_d(window->window, &mask, &window->style->white, rewind_xpm);
    playicon = gdk_pixmap_create_from_xpm_d(window->window, &mask, &window->style->white, play_xpm);
    pauseicon = gdk_pixmap_create_from_xpm_d(window->window, &mask, &window->style->white, pause_xpm);
    stopicon = gdk_pixmap_create_from_xpm_d(window->window, &mask, &window->style->white, stop_xpm);
    fastforwardicon = gdk_pixmap_create_from_xpm_d(window->window, &mask, &window->style->white, fastforward_xpm);

    iconwidget = gtk_pixmap_new(rewindicon, mask);
    button = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar), "Rewind", "Previous Track", "Private", iconwidget, GTK_SIGNAL_FUNC(buttonhandler), (gpointer) "rewind");
    gtk_toolbar_append_space(GTK_TOOLBAR(toolbar));
    iconwidget = gtk_pixmap_new(playicon, mask);
    button = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar), "Play", "Play", "Private", iconwidget, GTK_SIGNAL_FUNC(buttonhandler), (gpointer) "play");
    gtk_toolbar_append_space(GTK_TOOLBAR(toolbar));
    iconwidget = gtk_pixmap_new(pauseicon, mask);
    button = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar), "Pause", "Pause", "Private", iconwidget, GTK_SIGNAL_FUNC(buttonhandler), (gpointer) "pause");
    gtk_toolbar_append_space(GTK_TOOLBAR(toolbar));
    iconwidget = gtk_pixmap_new(stopicon, mask);
    button = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar), "Stop", "Stop", "Private", iconwidget, GTK_SIGNAL_FUNC(buttonhandler), (gpointer) "stop");
    gtk_toolbar_append_space(GTK_TOOLBAR(toolbar));
    iconwidget = gtk_pixmap_new(fastforwardicon, mask);
    button = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar), "FastForward", "Next Track", "Private", iconwidget, GTK_SIGNAL_FUNC(buttonhandler), (gpointer) "fastforward");
    gtk_box_pack_start(GTK_BOX(horizontal_toolbar_box), toolbar, FALSE, FALSE, 0);
    gtk_widget_show(toolbar);

    // Build toolbar for modifying the playlist
    toolbar = gtk_toolbar_new(GTK_ORIENTATION_HORIZONTAL, GTK_TOOLBAR_ICONS);
    gtk_container_set_border_width(GTK_CONTAINER(toolbar), 5);
    gtk_toolbar_set_space_size(GTK_TOOLBAR(toolbar), 5);

    addicon = gdk_pixmap_create_from_xpm_d(window->window, &mask, &window->style->white, add_xpm);
    removeicon = gdk_pixmap_create_from_xpm_d(window->window, &mask, &window->style->white, remove_xpm);
    moveupicon = gdk_pixmap_create_from_xpm_d(window->window, &mask, &window->style->white, moveup_xpm);
    movedownicon = gdk_pixmap_create_from_xpm_d(window->window, &mask, &window->style->white, movedown_xpm);

    iconwidget = gtk_pixmap_new(addicon, mask);
    button = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar), "Add", "Add a new item to the playlist", "Private", iconwidget, GTK_SIGNAL_FUNC(cb_add), NULL);
    gtk_toolbar_append_space(GTK_TOOLBAR(toolbar));
    iconwidget = gtk_pixmap_new(removeicon, mask);
    button = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar), "Remove", "Remove an item from the playlist", "Private", iconwidget, GTK_SIGNAL_FUNC(buttonhandler), (gpointer) "remove");
    gtk_toolbar_append_space(GTK_TOOLBAR(toolbar));
    iconwidget = gtk_pixmap_new(moveupicon, mask);
    button = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar), "Move_Up", "Move an item up in the playlist", "Private", iconwidget, GTK_SIGNAL_FUNC(buttonhandler), (gpointer) "moveup");
    gtk_toolbar_append_space(GTK_TOOLBAR(toolbar));
    iconwidget = gtk_pixmap_new(movedownicon, mask);
    button = gtk_toolbar_append_item(GTK_TOOLBAR(toolbar), "Move_Down", "Move an item down in the playlist", "Private", iconwidget, GTK_SIGNAL_FUNC(buttonhandler), (gpointer) "movedown");
    gtk_box_pack_start(GTK_BOX(horizontal_toolbar_box), toolbar, FALSE, FALSE, 0);
    gtk_widget_show(toolbar);
    gtk_box_pack_end(GTK_BOX(vertical_box), horizontal_toolbar_box, FALSE, TRUE, 0);
    gtk_widget_show(horizontal_toolbar_box);

    // Add the filename display:
    filename_label = gtk_label_new("Filename:");
    filename_textbox = gtk_entry_new();
    gtk_signal_connect(GTK_OBJECT(filename_textbox), "changed", GTK_SIGNAL_FUNC(cb_text_entry), &filename_container);
    filename_box = gtk_hbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(filename_box), filename_label, FALSE, FALSE, 2);
    gtk_box_pack_end(GTK_BOX(filename_box), filename_textbox, TRUE, TRUE, 0);
    gtk_widget_show(filename_label);
    gtk_widget_show(filename_textbox);
    gtk_box_pack_start(GTK_BOX(vertical_box), filename_box, FALSE, FALSE, 0);
    gtk_widget_show(filename_box);

    // Add the timer
    new_timer(mytimer);
#if LOCATION == END    
    gtk_box_pack_end(GTK_BOX(vertical_box), BOX(mytimer), FALSE, FALSE, 0);
#else
    gtk_box_pack_start(GTK_BOX(vertical_box), BOX(mytimer), FALSE, FALSE, 0);
#endif
    gtk_widget_show(BOX(mytimer));
    
    // Build the playlist area
    playlistwin = gtk_scrolled_window_new(NULL, NULL);
    gtk_container_set_border_width(GTK_CONTAINER(playlistwin), 0);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(playlistwin), GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);

    playlist = gtk_clist_new(1);
    gtk_clist_set_selection_mode(GTK_CLIST(playlist), GTK_SELECTION_SINGLE);
    gtk_clist_set_shadow_type(GTK_CLIST(playlist), GTK_SHADOW_OUT);
    gtk_clist_set_column_width(GTK_CLIST(playlist), 0, 200);
    gtk_clist_column_titles_hide(GTK_CLIST(playlist));
    gtk_container_add(GTK_CONTAINER(playlistwin), playlist);
    gtk_signal_connect(GTK_OBJECT(playlist), "select_row", GTK_SIGNAL_FUNC(clist_sel), NULL);
    gtk_signal_connect(GTK_OBJECT(playlist), "unselect_row", GTK_SIGNAL_FUNC(clist_unsel), NULL);
    current_row = -1;
    gtk_box_pack_start(GTK_BOX(vertical_box), playlistwin, TRUE, TRUE, 0);
    gtk_widget_show(playlistwin);
    gtk_widget_show(playlist);
    
    gtk_widget_show(vertical_box);
    // Finished building the GUI, now display the whole thing
    gtk_widget_show(window);

    // Load initial playlist from command-line (if any)
    for (i = 1; i < argc; i++)
    {
	if(strcmp(argv[i], "-h") == 0 || strcmp(argv[i], "--help") == 0 || strcmp(argv[i], "-?") == 0) {
	  g_print("PlaySPC_GTK v%s\nCopyright (C) 2002, JJ Foote w/ parts by Michael Thornbury\n"
                  "Distributed under the terms of the GPL.\n\n", CURRENT_VERSION);
	  g_print("Usage:\n  %s [-P] [-s] [-p port] [-b bitrate] [-S] "/* [-o] */ "[-e] [-B] [playlist/rar]\n\n", argv[0]);
          g_print("Options:\n");
	  g_print("-P\t\tAuto play.\n");
          g_print("-s\t\tEnables shoutcast server (mono, 64kbits, port 8000 by default)\n");
          g_print("-p port\t\tDefines the port used by the shoutcast server. Quickly killing\n"
                  "\t\tand restarting playspc_gtk with active listeners tends to\n"
                  "\t\trequire changing the port\n");
          g_print("-b bitrate\tDefine the shoutcast output bitrate. Handled entirely by the\n"
                  "\t\tencoder. Incorrect settings will spill out errors by the\n"
                  "\t\tencoder and will go ignored by playspc_gtk\n");
          g_print("-S\t\tTurns on stereo for shoutcast server\n");
/*         g_print("-o\t\tDisables sound output\n"); */
	  g_print("-e\t\tCauses output to be echoed.\n");
	  g_print("-B\t\tCauses output to be boosted.\n");
	  return 1;
	} else if (strcmp(argv[i], "-s") == 0) {
	  shoutcast = 1;
	} else if (strcmp(argv[i], "-S") == 0) {
	  stereo = 1;
/*	} else if (strcmp(argv[i], "-o") == 0) {
	  linuxout = -2; */
	} else if (strcmp(argv[i], "-b") == 0) {
          if (atoi(argv[++i]) != 0) {
             bitrate = atoi(argv[i]);
	     buf_size = bitrate*1024/8;
	  }
	} else if (strcmp(argv[i], "-B") == 0) {
	  boost = 1;
	  g_print("Boosting.\n");
	} else if (strcmp(argv[i], "-e") == 0) {
	  echo = 1;
	  g_print("Echoing.\n");
	} else if (strcmp(argv[i], "-p") == 0) {
          port = atoi(argv[++i]);
	} else if (strcmp(argv[i], "-P") == 0) {
          if (atoi(argv[i+1]) != 0)
             autoplay = atoi(argv[++i]);
	} else {
            open_multi_file_container(argv[i]);
	}
    }

    // Setup the encoder if need be.
    init2();

    // We're done now, so enter the GTK main loop and wait for the user to do something
    gdk_threads_enter(); // GTK is not thread-safe; lock GTK for this thread
    if (autoplay) {
       gtk_clist_select_row(GTK_CLIST(playlist), autoplay - 1, 0);
       play();
    }
    gtk_main();
    gdk_threads_leave();
    return 0;
}
