///////////////////////////////////////////////////////////////////////////
// JJ Foote w\ parts by Michael Thornbury                                //
// PlaySPC_GTK                                                           //
// Version 0.2.7 - February 21, 2003                                     //
//                                                                       //
// This file, spccore.c, was adapted with the author's permission from   //
// LITE.C, from OpenSPC source ver.300 for Linux.  The code has been     //
// stripped to what is needed to have a continuously running plugin.     //
// As a result, CLIP, and others have to be cleared between SPC files,   //
// so a new routine was added to sound.c, SNDFree.  This code should be  //
// safe from any obvious memory leaks guaranteeing a higher level of     //
// assurance it won't crash while streaming for an extended period of    //
// time.                                                                 //
//                                                                       //
// Please see the included 'openspc.txt' file for documentation and      //
// license information for code taken from OpenSPC.                      //
///////////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <fcntl.h>
#include <string.h>
#include "shared.h"
#include "openspc.h"

#define MAX_STRING_SIZE	200

int spc_play();
void Exit(int);
void Reset_SPC();
void* input_plugin_handler();

// Activities
#define NONE 0
#define PLAY 1
#define STOP 2
#define PAUSE 3
//#define NOT_RECEIVED 4
#define EXIT 255

#define BUFSIZE 32000

int freq = FREQ;
int activity = NONE;

pthread_mutex_t activity_mutex; // Checks for activities.
pthread_cond_t activity_cond;   // To tell others that an activity has occured.

char file[MAX_STRING_SIZE];
int limit;
char file_new[MAX_STRING_SIZE];
int limit_new;

void Update(void)
{
    const char data[COMMAND_SIZE] = "data";
    void *buf = NULL;
    int total;

    buf = malloc(BUFSIZE);
    total = OSPC_Run(-1, buf, BUFSIZE);

    // Pipe music to player.
    write(STDOUT_FILENO,data,COMMAND_SIZE);
    write(STDOUT_FILENO,&total,sizeof(int));
    write(STDOUT_FILENO, buf, total);
}

int main(int argc, char* argv[])
{
    int done=0;

    // A child thread (not too important to store).
    pthread_t child;
    pthread_attr_t attr;

    if (argc != 1) {
	fprintf(stderr,"Plugin called improperly.\n");
	return 1;
    }

    // Initialize activity mutex.
    pthread_mutex_init(&activity_mutex, NULL);

    pthread_cond_init(&activity_cond, NULL);

    pthread_attr_init(&attr);
    // Make it detached.
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
    // And one-to-one with kernel threads.
    pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);

    // Creates a seperate thread to handle input from the spc player.
    pthread_create(&child, &attr, input_plugin_handler, NULL);

    // Handle a spontaneous stop.
    signal(SIGPIPE, Exit);

    while(!done) {
	pthread_mutex_lock(&activity_mutex);

	// All activities should set activity to NONE before continuing
	// into the main loop.
	while (activity == NONE)
	    pthread_cond_wait(&activity_cond, &activity_mutex);

	switch (activity)
	{
	    case PLAY:
		activity = NONE;
		snprintf(file, MAX_STRING_SIZE, "%s", file_new);
		limit = limit_new;
		pthread_mutex_unlock(&activity_mutex);
		if (spc_play())
		   done = 1;
		break;
	    // Silly person keeps hitting stop when we're not playing.
	    case STOP:
		activity = NONE;
		pthread_mutex_unlock(&activity_mutex);
		break;
	    case EXIT:
		activity = NONE;
		pthread_mutex_unlock(&activity_mutex);
		done = 1;
		break;
	    default:
		fprintf(stderr, "Activity %d unknown.", activity);
		activity = NONE;
		pthread_mutex_unlock(&activity_mutex);
	};
    }
    return 0;
}

int spc_play()
{
    int seconds = 0,finished = 0,stopped = 0, urate = 0;
    const char done[COMMAND_SIZE] = "done";
    const char time[COMMAND_SIZE] = "time";
    int fd;
    off_t size;
    void *ptr;

    while (!stopped)
    {
	fd=open(file, O_RDONLY);
	if(fd < 0) {
	    fprintf(stderr, "\nUnable to open file '%s'!\n", file);
	    exit(1);
	}
	size = lseek(fd, 0, SEEK_END);
	lseek(fd, 0, SEEK_SET);
	ptr = malloc(size);
	read(fd, ptr, size);
	close(fd);

	if((fd=OSPC_Init(ptr, size))) {
	    fprintf(stderr, "\nOSPC_Init returned %d!\n", fd);
	    exit(1);
	}
	free(ptr);

	while (!finished)
	{
	    pthread_mutex_lock(&activity_mutex);

	    while (activity == PAUSE)
	    //       while ((activity == PAUSE) || (activity == NOT_RECEIVED))
		pthread_cond_wait(&activity_cond, &activity_mutex);

	    // Only send the next chunk if we've unpaused or a request for the next chunk has been sent.
	    if (activity == NONE)
		Update();
	    else {
	       // Special cases to break out of loop
	       if (activity == STOP) {
		  activity = NONE;
		  stopped = 1;
		  seconds = 0;
		  pthread_mutex_unlock(&activity_mutex);
		  break;
	       } else if (activity == PLAY) {
		  activity = NONE;
		  snprintf(file, MAX_STRING_SIZE, "%s", file_new);
		  limit = limit_new;
		  seconds = 0;
		  pthread_mutex_unlock(&activity_mutex);
		  break;
	       }
	       switch (activity)
	       {
		  case PAUSE: // do nothing
		    break;
		  case EXIT:
		    activity = NONE;
		    pthread_mutex_unlock(&activity_mutex);
		    return 1;
		  default:
		    fprintf(stderr, "Activity %d unknown.", activity);
		    activity = NONE;
	       };
	    }

	    /* This is an ugly solution (we assume that it takes
	     * exactly one second to play 4 buffers full of sound).
	     * Ideally, libopenspc should have a function that reports
	     * the time elapsed, or some easy way to calculate it.
	     */
	    urate++;
	    if(urate == 4) {
		urate = 0;
		seconds++;
		write(STDOUT_FILENO, time, COMMAND_SIZE);
		write(STDOUT_FILENO, &seconds, sizeof(int));

		if (limit) {
		     if (seconds==limit) {
			stopped = 1;
			finished = 1;
		     }
		}
	    }

	    //     activity == NOT_RECEIVED;
	    pthread_mutex_unlock(&activity_mutex);
	}
    }
    // Only report done when it's a non-forced exit.
    if (finished)
	write(STDOUT_FILENO, done, COMMAND_SIZE);
    return(0);
}

void *input_plugin_handler()
{
    char command[COMMAND_SIZE+1];

    command[COMMAND_SIZE] = '\0';

    while(1) {
	// Wait for input.
	while (read(STDIN_FILENO, command, COMMAND_SIZE) <= 0);

	pthread_mutex_lock(&activity_mutex);

	// Set the activity variable based on the activity to do.
	if (strcmp(command, "play") == 0) {
	    read(STDIN_FILENO, &limit_new, sizeof(int));
	    read(STDIN_FILENO, file_new, MAX_STRING_SIZE);
	    activity = PLAY;
	}
	else if (strcmp(command, "stop") == 0)
	    activity = STOP;
	else if (strcmp(command, "freq") == 0) {
		 read(STDIN_FILENO, &freq, sizeof(int));
		 activity = NONE;
	} else if (strcmp(command, "pause") == 0)
		 activity = PAUSE;
	else if (strcmp(command, "unpause") == 0)
		 activity = NONE;
	/* else if (strcmp(command, "send") == 0) {
		 if (activity == NOT_RECEIVED)
		    activity = NONE;
	} */ else if (strcmp(command, "exit") == 0)
		 activity = EXIT;

	pthread_cond_signal(&activity_cond);

	pthread_mutex_unlock(&activity_mutex);
    }
}

void Exit(int x)
{
    exit(0);
}
